# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for analyzing spatially embedded complex networks, handling
multivariate data and generating time series surrogates.
"""

import numpy as np
from numpy import random

from ._ext.types import to_cy, ADJ, NODE, DWEIGHT, DFIELD
from ._ext.numerics import _randomlySetCrossLinks, _randomlyRewireCrossLinks, \
    _cross_transitivity, _nsi_cross_transitivity, _cross_local_clustering, \
    _nsi_cross_local_clustering

from .network import Network, NetworkError


class InteractingNetworks(Network):

    """
    Encapsulates an ensemble of interacting networks.

    Provides measures to analyze the interaction topology of different pairs of
    subnetworks (groups of vertices).

    So far, most methods only give meaningful results for undirected networks!

    The idea of interacting networks and measures for their analysis are
    described in [Donges2011a]_.

    Consistently node-weighted measures for interacting network topologies are
    derived, described and applied in [Wiedermann2011]_.
    """
    #
    #  Definitions of internal methods
    #

    def __init__(self, adjacency, directed=False, node_weights=None,
                 silence_level=0):
        """
        Initialize an instance of InteractingNetworks.

        :type adjacency: square numpy array or list [node,node] of 0s and 1s
        :arg  adjacency: Adjacency matrix of the new network.  Entry [i,j]
            indicates whether node i links to node j.  Its diagonal must be
            zero.  Must be symmetric if directed=False.
        :arg bool directed: Indicates whether the network shall be considered
            as directed. If False, adjacency must be symmetric.
        :type node_weights: 1d numpy array or list [node] of floats >= 0
        :arg  node_weights: Optional array or list of node weights to be used
            for node splitting invariant network measures.  Entry [i] is the
            weight of node i.  (Default: list of ones)
        :arg int silence_level: The inverse level of verbosity of the object.
        """
        Network.__init__(self, adjacency=adjacency, directed=directed,
                         node_weights=node_weights,
                         silence_level=silence_level)

    def __str__(self):
        """
        Return a string representation of InteractingNetworks object.
        """
        return 'InteractingNetworks:\n' + Network.__str__(self)

    #
    #  Graph generation methods
    #

    @staticmethod
    def SmallTestNetwork():
        """
        Return a 6-node undirected test network.

        The network looks like this::

                3 - 1
                |   | \\
            5 - 0 - 4 - 2

        :rtype: InteractingNetworks instance
        :return: an InteractingNetworks instance for testing purposes.
        """
        nw = InteractingNetworks(adjacency=[[0, 0, 0, 1, 1, 1],
                                            [0, 0, 1, 1, 1, 0],
                                            [0, 1, 0, 0, 1, 0],
                                            [1, 1, 0, 0, 0, 0],
                                            [1, 1, 1, 0, 0, 0],
                                            [1, 0, 0, 0, 0, 0]],
                                 directed=False,
                                 node_weights=[0.6, 0.8, 1.0, 1.2, 1.4, 1.6],
                                 silence_level=2)
        link_weights = np.array([[0, 0, 0, 1.3, 2.5, 1.1],
                                 [0, 0, 2.3, 2.9, 2.7, 0],
                                 [0, 2.3, 0, 0, 1.5, 0],
                                 [1.3, 2.9, 0, 0, 0, 0],
                                 [2.5, 2.7, 1.5, 0, 0, 0],
                                 [1.1, 0, 0, 0, 0, 0]])
        nw.set_link_attribute("link_weights", link_weights)
        return nw

    @staticmethod
    def SmallDirectedTestNetwork():
        """
        Return a 6-node directed test network with node and edge weights.

        The node weights are [1.5, 1.7, 1.9, 2.1, 2.3, 2.5],
        a typical node weight for corrected n.s.i. measures would be 2.0.

        :rtype: InteractingNetworks instance
        """
        nw = InteractingNetworks(adjacency=[[0, 1, 0, 1, 0, 0],
                                            [0, 0, 1, 0, 1, 0],
                                            [0, 0, 0, 0, 0, 0],
                                            [0, 1, 0, 0, 0, 0],
                                            [1, 0, 1, 0, 0, 0],
                                            [1, 0, 0, 0, 0, 0]],
                                 directed=True,
                                 node_weights=[0.6, 0.8, 1.0, 1.2, 1.4, 1.6],
                                 silence_level=1)
        nw.set_link_attribute("link_weights", np.array([[0, 1.3, 0, 2.5, 0, 0],
                                                        [0, 0, 1.9, 0, 1.0, 0],
                                                        [0, 0, 0, 0, 0, 0],
                                                        [0, 3.0, 0, 0, 0, 0],
                                                        [2.1, 0, 2.7, 0, 0, 0],
                                                        [1.5, 0, 0, 0, 0, 0]]))
        return nw

    @staticmethod
    def RandomlySetCrossLinks(network, node_list1, node_list2,
                              cross_link_density=None,
                              number_cross_links=None):
        """
        Creates a set of random cross links between the considered
        interacting subnetworks. The number of cross links to be set
        can be chosen either explicitly or via a predefined cross link density.
        By not choosing any of either, a null model is created under
        preservation of the cross link density of the initial network.

        Implementation:

        Determines the number of cross links to be set.
        Creates an empty cross adjacency matrix.
        Randomly picks the coordinates of an entry and sets it to one.
        Repeats the procedure until the desired cross link density
        is reached.

        :type network: InteractingNetworks instance
        :arg network: The base network for setting random cross links.
        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :rtype:  :class:`InteractingNetworks`
        :return: The initial InteractingNetworks with random cross links
        """
        #  store node lists as arrays
        nodes1 = np.array(node_list1, dtype=NODE)
        nodes2 = np.array(node_list2, dtype=NODE)
        #  retrieve number of nodes
        N1, N2 = len(nodes1), len(nodes2)
        #  retrieve cross adjacency matrix
        cross_A = network.cross_adjacency(nodes1, nodes2).astype(ADJ)

        #  determine number of cross links
        if cross_link_density is not None:
            number_cross_links = int(cross_link_density * (N1 * N2))
            print("Setting number of cross links according to "
                  "chosen link density.")
        elif cross_link_density is None and number_cross_links is None:
            number_cross_links = int(cross_A.sum())
            print("Creating a null model for the given interacting networks.")
        #  else: take the explicitly chosen number of cross links
        if number_cross_links > (N1 * N2):
            print("The number of cross links exceeds maximum.")
            print("Setting link density of initial interacting network.")
            number_cross_links = int(cross_A.sum())

        #  retrieve adjacency matrix of the full interacting network
        A_new = network.adjacency.astype(ADJ)
        #  create new empty cross adjacency matrix
        cross_A_new = np.zeros((N1, N2), dtype=ADJ)

        _randomlySetCrossLinks(A_new, cross_A_new, number_cross_links,
                               nodes1, nodes2, N1, N2)
        return InteractingNetworks(adjacency=A_new,
                                   directed=network.directed,
                                   node_weights=network.node_weights,
                                   silence_level=network.silence_level)

    @staticmethod
    def RandomlySetCrossLinks_sparse(network, node_list1, node_list2,
                                     cross_link_density=None,
                                     number_cross_links=None):
        """
        Creates a set of random cross links between the considered
        interacting subnetworks. The number of cross links to be set
        can be chosen either explicitly or via a predefined cross link density.
        By not choosing any of either, a null model is created under
        preservation of the cross link density of the initial network.

        Implementation:

        Determines the number of cross links to be set.
        Creates an empty cross adjacency matrix.
        Randomly picks the coordinates of an entry and sets it to one.
        Repeats the procedure until the desired cross link density
        is reached.

        :type network: InteractingNetworks instance
        :arg network: The base network for setting random cross links.
        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :rtype:  :class:`InteractingNetworks`
        :return: The initial InteractingNetworks with random cross links
        """
        #  store node lists as arrays
        nodes1 = np.array(node_list1, dtype=int)
        nodes2 = np.array(node_list2, dtype=int)
        #  retrieve number of nodes
        N1, N2 = len(nodes1), len(nodes2)
        #  retrieve cross adjacency matrix
        cross_A = network.cross_adjacency_sparse(nodes1, nodes2).astype(int)

        #  determine number of cross links
        if cross_link_density is not None:
            number_cross_links = int(cross_link_density * (N1 * N2))
            print("Setting number of cross links according to chosen \
                  link density.")
        elif cross_link_density is None and number_cross_links is None:
            number_cross_links = int(sum(cross_A.values()))
            print("Creating a null model for the given interacting networks.")
        #  else: take the explicitly chosen number of cross links

        if number_cross_links > (N1 * N2):
            print("The number of cross links exceeds maximum.")
            print("Setting link density of initial interacting network.")
            number_cross_links = int(sum(cross_A.values()))

        #  retrieve adjacency matrix of the full interacting network
        A_new = network.sp_A.astype(int)
        #  create new empty cross adjacency matrix
        cross_A_new = np.zeros((N1, N2))

        n_1, n_2 = 0, 0
        node1, node2 = 0, 0
        for i in range(number_cross_links):
            while True:
                n_1 = int(random.random() * N1)
                n_2 = int(random.random() * N2)
                if not (cross_A_new[n_1, n_2] == 1):
                    break
            cross_A_new[n_1, n_2] = 1
        for i in range(N1):
            for j in range(N2):
                node1 = int(nodes1[i])
                node2 = int(nodes2[j])
                A_new[node1, node2] = cross_A_new[i, j]
                A_new[node2, node1] = cross_A_new[i, j]

        return InteractingNetworks(adjacency=A_new,
                                   directed=network.directed,
                                   node_weights=network.node_weights,
                                   silence_level=network.silence_level)

    @staticmethod
    def RandomlyRewireCrossLinks(network, node_list1, node_list2, swaps):
        """
        Randomize the cross links between two subnetworks under preservation of
        cross degree centrality of both subnetworks.

        Chooses randomly two cross links and swaps their ending points in
        subnetwork 2.

        Implementation:

        Stores the coordinates of the "1"-entries of the cross adjacency matrix
        in a tuple. Chooses randomly two entries of the tuple (ergo two cross
        links) allowing for the constraints that

        (1) the chosen links have distinct starting points in subnetwork 1 and
            distinct ending points in subnetwork 2
        (2) there do not exist intermediate links such that starting point of
            link 1 is connected to ending point of link 2 and vice versa.

        [In case two links have the same starting point or / and the same
        ending point, condition (2) is never satisfied. Therefore only
        condition (2) is implemented.]

        Swaps the ending points of the links in subnetwork 2 and overwrites the
        coordinates of the initial links in the tuple. The number of
        permutation procedures is determined by the "swaps" argument and the
        initial number of cross links. Creates a new adjacency matrix out of
        the altered tuple of coordinates.

        **Example** (Degree and cross degree sequences should be the same after
        rewiring):

        >>> net = InteractingNetworks.SmallTestNetwork()
        >>> print("Degree:", net.degree())
        Degree: [3 3 2 2 3 1]
        >>> print("Cross degree:", net.cross_degree(
        ...     node_list1=[0,3,5], node_list2=[1,2,4]))
        Cross degree: [1 1 0]
        >>> rewired_net = net.RandomlyRewireCrossLinks(
        ...     network=net, node_list1=[0,3,5],
        ...     node_list2=[1,2,4], swaps=10.)
        >>> print("Degree:", rewired_net.degree())
        Degree: [3 3 2 2 3 1]
        >>> print("Cross degree:", rewired_net.cross_degree(
        ...     node_list1=[0,3,5], node_list2=[1,2,4]))
        Cross degree: [1 1 0]

        :type network: :class:`InteractingNetworks` instance
        :arg network: The base network for rewiring cross links.
        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :arg float internal: Gives the fraction number_swaps /
            number_cross_links.
        :rtype:  :class:`InteractingNetworks`
        :return: The initial InteractingNetworks with swapped cross links
        """
        #  retrieve cross adjacency matrix of the considered interacting
        #  network
        cross_A = network.cross_adjacency(node_list1, node_list2).astype(ADJ)
        #  determine number of cross links
        number_cross_links = cross_A.sum()
        #  Store node lists as arrays
        nodes1 = np.array(node_list1, dtype=NODE)
        nodes2 = np.array(node_list2, dtype=NODE)
        #  retrieve adjacency matrix of the full interacting network
        A_new = network.adjacency.astype(ADJ)
        #  determine number of cross link permutations that will be performed
        number_swaps = NODE(swaps * number_cross_links)
        #  Create list of cross links
        cross_links = np.array(cross_A.nonzero(), dtype=NODE).transpose()

        _randomlyRewireCrossLinks(A_new, cross_A, cross_links, nodes1, nodes2,
                                  number_cross_links, number_swaps)
        return InteractingNetworks(adjacency=A_new,
                                   directed=network.directed,
                                   node_weights=network.node_weights,
                                   silence_level=network.silence_level)

    #
    #  Define methods for handling the interacting networks
    #

    def subnetwork(self, node_list):
        """
        Return the subnetwork induced by a subset of nodes as a Network object.

        This can be used to conveniently analyze the subnetwork separately,
        e.g., for calculation network measures solely this subnetwork.

        :arg [int] node_list: list of node indices describing the subnetwork
        :rtype: Network
        :return: the subnetwork induced by the nodes in node_list.
        """
        return Network(adjacency=self.internal_adjacency(node_list),
                       directed=self.directed,
                       node_weights=self.node_weights[node_list],
                       silence_level=self.silence_level)

    def internal_adjacency(self, node_list):
        """
        Return the adjacency matrix of a subnetwork induced by a subset of
        nodes.

        **Examples:**

        >>> InteractingNetworks.SmallTestNetwork().\
                internal_adjacency([0,3,5])
        array([[0, 1, 1], [1, 0, 0], [1, 0, 0]], dtype=int8)
        >>> InteractingNetworks.SmallTestNetwork().\
                internal_adjacency([1,2,4])
        array([[0, 1, 1], [1, 0, 1], [1, 1, 0]], dtype=int8)

        :arg [int] node_list: list of node indices describing the subnetwork
        :rtype: 2D array [node index, node index]
        :return: the subnetwork's adjacency matrix.
        """
        #  Create igraph Graph object describing the subgraph
        subgraph = self.graph.subgraph(node_list)
        #  Get adjacency matrix
        return np.array(subgraph.get_adjacency(type=2).data).astype(np.int8)

    def cross_adjacency(self, node_list1, node_list2):
        """
        Return cross adjacency matrix describing the interaction of two
        subnetworks.

        The cross adjacency matrix entry :math:`CA_{ij} = 1` describes that
        node i in the first subnetwork is linked to node j in the second
        subnetwork.  Vice versa, :math:`CA_{ji} = 1` indicates that node j in
        the first subnetwork is linked to node i in the second subnetwork.

        .. note::
           The Cross adjacency matrix is NEITHER square NOR symmetric in
           general!

        **Examples:**

        >>> r(InteractingNetworks.SmallTestNetwork().\
                cross_adjacency([1,2,4], [0,3,5]))
        array([[0, 1, 0], [0, 0, 0], [1, 0, 0]])
        >>> r(InteractingNetworks.SmallTestNetwork().\
                cross_adjacency([1,2,3,4], [0,5]))
        array([[0, 0], [0, 0], [1, 0], [1, 0]])

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :rtype: 2D array [node index_1, node index_2]
        :return: the cross adjacency matrix.
        """
        return self.adjacency[node_list1, :][:, node_list2]

    def cross_adjacency_sparse(self, node_list1, node_list2):
        """
        Return cross adjacency matrix describing the interaction of two
        subnetworks.

        The cross adjacency matrix entry M{CA_ij = 1} describes that node i
        in the first subnetwork is linked to node j in the second subnetwork.
        Vice versa, M{CA_ji = 1} indicates that node j in the first subnetwork
        is linked to node i in the second subnetwork.

        .. note::

           The Cross adjacency matrix is NEITHER square NOR symmetric in
           general!

        Examples:

        >>> print(InteractingNetworks.SmallTestNetwork().\
                cross_adjacency_sparse([1,2,4], [0,3,5]))
        [[0 1 0] [0 0 0] [1 0 0]]

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :rtype: 2D array [node index_1, node index_2]
        :return: the cross adjacency matrix.
        """
        return self.sp_A[node_list1, :][:, node_list2].A

    def internal_link_attribute(self, attribute_name, node_list):
        """
        Return a like attribute matrix of a subnetwork induced by a subset of
        nodes.

        **Example:**

        >>> net = InteractingNetworks.SmallTestNetwork()
        >>> r(net.internal_link_attribute("link_weights", [1,2,3]))
        array([[ 0. ,  2.3,  2.9],
               [ 2.3,  0. ,  0. ],
               [ 2.9,  0. ,  0. ]])

        :arg str attribute_name: _name of link attribute to be used
        :arg [int] node_list: list of node indices describing the subnetwork
        :rtype: square numpy array [node_index, node_index]
        :return: link weights submatrix
        """
        weights = np.zeros((len(node_list), len(node_list)))
        subgraph = self.graph.subgraph(node_list)

        if self.directed:
            for e in subgraph.es:
                weights[e.tuple] = e[attribute_name]
        #  Symmetrize if subgraph is undirected
        else:
            for e in subgraph.es:
                weights[e.tuple] = e[attribute_name]
                weights[e.tuple[1], e.tuple[0]] = e[attribute_name]

        return weights

    def cross_link_attribute(self, attribute_name, node_list1, node_list2):
        """
        Return a cross link weights matrix describing the interaction of two
        subnetworks.

        The cross link weights matrix entry :math:`CW_{ij} = w` describes that
        node i in the first subnetwork is linked to node j in the second
        subnetwork with weights :math:`w`.

        .. note::

           The cross link weights matrix is NEITHER square NOR symmetric in
           general!

        Example:

        >>> net = InteractingNetworks.SmallTestNetwork()
        >>> r(net.cross_link_attribute("link_weights", [1,2,3], [0,4]))
        array([[ 0. ,  2.7],
               [ 0. ,  1.5],
               [ 1.3,  0. ]])

        :arg str attribute_name: _name of link attribute to be used
        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :rtype: 2D array [node index_1, node index_2]
        :return: the cross adjacency matrix.
        """
        W = self.link_attribute(attribute_name)
        return W[node_list1, :][:, node_list2]

    def internal_path_lengths(self, node_list, link_attribute=None):
        """
        Return internal path length matrix of an induced subnetwork.

        Contains the paths length between all pairs of nodes within the
        subnetwork. However, the paths themselves will generally contain nodes
        from the full network. To avoid this and only consider paths lying
        within the subnetwork, do the following:

        >>> InteractingNetworks.SmallTestNetwork().\
                subnetwork([0,3,5]).path_lengths()
        array([[ 0., 1., 1.], [ 1., 0., 2.], [ 1., 2., 0.]])

        **Examples:**

        >>> InteractingNetworks.SmallTestNetwork().\
                internal_path_lengths([0,3,5], None)
        array([[ 0., 1., 1.], [ 1., 0., 2.], [ 1., 2., 0.]])
        >>> InteractingNetworks.SmallTestNetwork().\
                internal_path_lengths([1,2,4], None)
        array([[ 0., 1., 1.], [ 1., 0., 1.], [ 1., 1., 0.]])

        :arg [int] node_list: list of node indices describing the subnetwork
        :arg str link_attribute: Optional name of the link attribute to be used
            as the links' length. If None, links have length 1. (Default: None)
        :rtype: 2D array [node index, node index]
        :return: the internal path length matrix of an induced subnetwork.
        """
        return self.\
            path_lengths(link_attribute)[node_list, :][:, node_list]

    def cross_path_lengths(self, node_list1, node_list2, link_attribute=None):
        """
        Return cross path length matrix for a pair of subnetworks.

        Contains the path length between nodes from different subnetworks. The
        paths may generally contain nodes from the full network.

        **Examples:**

        >>> InteractingNetworks.SmallTestNetwork().\
                cross_path_lengths([0,3,5], [1,2,4], None)
        array([[ 2.,  2.,  1.], [ 1.,  2.,  2.], [ 3.,  3.,  2.]])
        >>> InteractingNetworks.SmallTestNetwork().\
                cross_path_lengths([0,5], [1,2,3,4], None)
        array([[ 2.,  2.,  1.,  1.], [ 3.,  3.,  2.,  2.]])

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :arg str link_attribute: Optional name of the link attribute to be used
            as the links' length. If None, links have length 1. (Default: None)
        :rtype: 2D array [index1, index2]
        :return: the cross path length matrix for a pair of subnetworks.
        """
        return self.path_lengths(link_attribute)[node_list1, :][:, node_list2]

    #
    #  Define scalar statistics for interacting networks
    #

    def number_cross_links(self, node_list1, node_list2):
        """
        Return the number of links connecting the two subnetworks.

        **Examples:**

        >>> InteractingNetworks.SmallTestNetwork().\
                number_cross_links([0,3,5], [1,2,4])
        2
        >>> InteractingNetworks.SmallTestNetwork().\
                number_cross_links([0,5], [1,2,3,4])
        2

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :return int: the number of links between nodes from different
            subnetworks.
        """
        if self.directed:
            raise NetworkError("Not implemented yet...")

        return self.cross_adjacency(node_list1, node_list2).sum()

    def total_cross_degree(self, node_list1, node_list2):
        """
        Return the total cross degree of the two subnetworks.

        **Examples:**

        >>> InteractingNetworks.SmallTestNetwork().\
                toal_cross_degree([0,3,5], [1,2,4])
        1.0
        >>> InteractingNetworks.SmallTestNetwork().\
                total_cross_degree([0,5], [1,2,3,4]).round(4)
        0.6667

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :return int: the total cross degree.
        """
        return np.mean(self.cross_degree(node_list1, node_list2))

    def number_internal_links(self, node_list):
        """
        Return the number of links within an induced subnetwork.

        **Examples:**

        >>> InteractingNetworks.SmallTestNetwork().\
                number_internal_links([0,3,5])
        2
        >>> InteractingNetworks.SmallTestNetwork().\
                number_internal_links([1,2,4])
        3

        :arg [int] node_list: list of node indices describing the subnetwork
        :return int: the number of links within a given subnetwork.
        """
        n_links = self.internal_adjacency(node_list).sum()
        if self.directed:
            return n_links
        else:
            return n_links // 2

    def cross_degree_density(self, node_list1, node_list2):
        """
        Return the density of degrees, i.e., the cross degree sequence of the
        first subnetwork normalized to the number of nodes in the second
        subnetwork

        **Example:**

        >>> InteractingNetworks.SmallTestNetwork().\
                cross_degree_density([0,3,5], [1,2,4])
        array([0.33333333, 0.33333333, 0.        ])

        :arg [int] node_list1: list of node indices describing the first
                               subnetwork

        :arg [int] node_list2: list of node indices describing the second
                               subnetwork

        :rtype: 1D array [node index]
        :return: the cross degree sequence.
        """

        N2 = len(node_list2)
        return self.cross_degree(node_list1, node_list2) / N2

    def cross_link_density(self, node_list1, node_list2):
        """
        Return the density of links between two subnetworks.

        **Examples:**

        >>> r(InteractingNetworks.SmallTestNetwork().\
                cross_link_density([0,3,5], [1,2,4]))
        0.2222
        >>> InteractingNetworks.SmallTestNetwork().\
                cross_link_density([0,5], [1,2,3,4])
        0.25

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :return float: the density of links between two subnetworks.
        """
        N1, N2 = len(node_list1), len(node_list2)
        if self.directed:
            raise NetworkError("Not implemented yet...")

        n_cl = self.number_cross_links(node_list1, node_list2)
        return float(n_cl) / (N1 * N2)

    def internal_link_density(self, node_list):
        """
        Return the density of links within an induced subnetwork.

        **Examples:**

        >>> r(InteractingNetworks.SmallTestNetwork().\
                internal_link_density([0,3,5]))
        0.6667
        >>> r(InteractingNetworks.SmallTestNetwork().\
                internal_link_density([1,2,3,4]))
        0.6667

        :arg [int] node_list: list of node indices describing the subnetwork
        :return float: the density of links within a subnetwork.
        """
        N = len(node_list)
        n_links = InteractingNetworks.number_internal_links(self, node_list)
        if self.directed:
            return float(n_links) / (N * (N - 1))
        else:
            return 2 * float(n_links) / (N * (N - 1))

    def internal_global_clustering(self, node_list):
        """
        Return internal global clustering coefficients for an induced
        subnetwork.

        Internal global clustering coefficients are calculated as mean values
        from the local clustering sequence of the whole network. This
        implies that triangles spanning different subnetworks will generally
        contribute to the internal clustering coefficient.

        To avoid this and consider only triangles lying within the subnetwork:

        >>> r(InteractingNetworks.SmallTestNetwork().\
                subnetwork([0,3,5]).global_clustering())
        0.0

        **Examples:**

        >>> r(InteractingNetworks.SmallTestNetwork().\
                internal_global_clustering([0,3,5]))
        0.0
        >>> r(InteractingNetworks.SmallTestNetwork().\
                internal_global_clustering([1,2,4]))
        0.5556

        :arg [int] node_list: list of node indices describing the subnetwork
        :return float: the internal global clustering coefficient for a
            subnetwork.
        """
        clustering = self.local_clustering()
        internal_clustering = clustering[node_list].mean()
        return internal_clustering

    def cross_global_clustering(self, node_list1, node_list2):
        """
        Return global cross clustering for a pair of subnetworks.

        The global cross clustering coefficient C_v gives the average
        probability, that two randomly drawn neighbors in subnetwork 2 of node
        v in subnetwork 1 are also neighbors and vice versa. It counts
        triangles having one vertex in subnetwork 1 and two vertices in
        subnetwork 2 and vice versa.

        **Examples:**

        >>> InteractingNetworks.SmallTestNetwork().\
                cross_global_clustering([0,3,5], [1,2,4])
        0.0
        >>> InteractingNetworks.SmallTestNetwork().\
                cross_global_clustering([2], [1,3,4])
        1.0
        >>> InteractingNetworks.SmallTestNetwork().\
                cross_global_clustering([3,4], [1,2])
        0.5

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :return float: the cross global clustering coefficient for a pair of
            subnetworks.
        """
        #  Get cross local clustering sequences
        cc = InteractingNetworks.cross_local_clustering(self,
                                                        node_list1, node_list2)
        return cc.mean()

    def cross_global_clustering_sparse(self, node_list1, node_list2):
        """
        Return global cross clustering for a pair of subnetworks.

        The global cross clustering coefficient C_v gives the average
        probability, that two randomly drawn neighbors in subnetwork 2 of node
        v in subnetwork 1 are also neighbors and vice versa. It counts
        triangles having one vertex in subnetwork 1 and two vertices in
        subnetwork 2 and vice versa.

        Examples:

        >>> InteractingNetworks.SmallTestNetwork().\
                cross_global_clustering_sparse([0,3,5], [1,2,4])
        0.0
        >>> InteractingNetworks.SmallTestNetwork().\
                cross_global_clustering_sparse([2], [1,3,4])
        1.0
        >>> InteractingNetworks.SmallTestNetwork().\
                cross_global_clustering_sparse([3,4], [1,2])
        0.5

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :return float: the cross global clustering coefficient for a pair of
            subnetworks.
        """
        #  Get cross local clustering sequences
        cc = self.cross_local_clustering_sparse(node_list1, node_list2)
        return cc.mean()

    def cross_transitivity(self, node_list1, node_list2):
        """
        Return cross transitivity for a pair of subnetworks.

        The cross transitivity is the probability, that two randomly drawn
        neighbors in subnetwork 2 of node v in subnetwork 1 are also neighbors.
        It counts triangles having one vertex in subnetwork 1 and two vertices
        in subnetwork 2. Cross transitivity tends to weight low cross degree
        vertices less strongly when compared to the global cross clustering
        coefficient (see [Newman2003]_).

        **Examples:**

        >>> InteractingNetworks.SmallTestNetwork().\
                cross_transitivity([0,3,5], [1,2,4])
        0.0
        >>> InteractingNetworks.SmallTestNetwork().\
                cross_transitivity([2], [1,3,4])
        1.0
        >>> InteractingNetworks.SmallTestNetwork().\
                cross_transitivity([3,4], [1,2])
        1.0

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :return float: the cross transitivity for a pair of subnetworks.
        """
        return _cross_transitivity(
            to_cy(self.adjacency, ADJ),
            np.array(node_list1, dtype=NODE), np.array(node_list2, dtype=NODE))

    def cross_transitivity_sparse(self, node_list1, node_list2):
        """
        Return cross transitivity for a pair of subnetworks.

        The cross transitivity is the probability, that two randomly drawn
        neighbors in subnetwork 2 of node v in subnetwork 1 are also
        neighbors. It counts triangles having one vertex in
        subnetwork 1 and two vertices in subnetwork 2. Cross
        transitivity tends to weight low cross degree vertices less strongly
        when compared to the global cross clustering coefficient (see Newman,
        SIAM Review, 2003).

        Examples:

        >>> InteractingNetworks.SmallTestNetwork().\
                cross_transitivity_sparse([0,3,5], [1,2,4])
        0.0
        >>> InteractingNetworks.SmallTestNetwork().\
                cross_transitivity_sparse([3,4], [1,2])
        1.0

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :return float: the cross transitivity for a pair of subnetworks.
        """
        cross_degree = self.cross_degree(node_list1, node_list2)

        #  Get sparse adjacency matrix
        A = self.sp_A[node_list1+node_list2, :][:, node_list1+node_list2]
        #  Get subnetwork sizes
        N1, N2 = len(node_list1), len(node_list2)
        #  Initialize
        cross_transitivity = 0.0
        #  Set counter
        counter_triangles = 0.0
        counter_triples = 0.0
        #  Calculate cross transitivity from subnetwork 1 to subnetwork 2
        #  Loop over nodes in subnetwork 1
        for i in range(N1):
            node1 = i
            if cross_degree[i] > 1:
                #  Loop over unique pairs of nodes in subnetwork 2
                for j in range(N1, N1+N2):
                    node2 = j
                    for k in range(N1, j):
                        node3 = k
                        if A[node1, node2] == 1 and A[node1, node3] == 1:
                            counter_triples += 1
                            if A[node2, node3] == 1:
                                counter_triangles += 1

        if counter_triples:
            cross_transitivity = counter_triangles / counter_triples
        return cross_transitivity

    @staticmethod
    def _calculate_general_average_path_length(path_lengths, internal=False):
        """
        Calculate general average path length for interacting networks.

        :type path_lengths: 2D array [index, index]
        :arg path_lengths: The path length matrix.
        :arg bool internal: Indicates, whether internal or cross average path
            length shall be calculated.
        :return float: the general average path length.
        """
        #  Get shape of path lengths array for normalization
        (N, M) = path_lengths.shape

        #  Identify unconnected pairs and save in binary array isinf
        unconnected_pairs = np.isinf(path_lengths)
        #  Count the number of unconnected pairs
        n_unconnected_pairs = unconnected_pairs.sum()
        #  Set infinite entries corresponding to unconnected pairs to zero
        path_lengths[unconnected_pairs] = 0

        #  Take average of shortest geographical path length matrix optionally
        #  excluding the diagonal, since it is always zero, and all unconnected
        #  pairs. The diagonal should never contain infinities, so that should
        #  not be a problem.
        if internal:
            norm = float((N - 1) * M - n_unconnected_pairs)
        else:
            norm = float(N * M - n_unconnected_pairs)

        average_path_length = path_lengths.sum() / norm

        #  Reverse changes to path_lengths
        path_lengths[unconnected_pairs] = np.inf
        return average_path_length

    def cross_average_path_length(self, node_list1, node_list2,
                                  link_attribute=None):
        """
        Return cross average path length.

        Return the average (weighted) shortest path length between two induced
        subnetworks.

        **Examples:**

        >>> InteractingNetworks.SmallTestNetwork().\
                cross_average_path_length([0,3,5], [1,2,4], None)
        2.0
        >>> InteractingNetworks.SmallTestNetwork().\
                cross_average_path_length([0,5], [1,2,3,4], None)
        2.0

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :arg str link_attribute: Optional name of the link attribute to be used
            as the links' length. If None, links have length 1. (Default: None)
        :return float: the cross average path length between a pair of
            subnetworks.
        """
        path_lengths = InteractingNetworks.cross_path_lengths(
            self, node_list1, node_list2, link_attribute)

        return self._calculate_general_average_path_length(
            path_lengths, internal=False)

    def internal_average_path_length(self, node_list, link_attribute=None):
        """
        Return internal average path length for an induced subnetwork.

        Return the average (weighted) shortest path length between all pairs
        of nodes within a subnetwork separately for which a path exists. Paths
        between nodes from different subnetworks are not included in the
        average!

        However, even if the end points lie within the same layer, the paths
        themselves will generally contain nodes from the whole network. To
        avoid this and only consider paths lying within the subnetwork, do the
        following:

        >>> r(InteractingNetworks.SmallTestNetwork().\
                subnetwork([0,3,5]).average_path_length(None))
        1.3333

        **Examples:**

        >>> r(InteractingNetworks.SmallTestNetwork().\
                internal_average_path_length([0,3,5], None))
        1.3333
        >>> r(InteractingNetworks.SmallTestNetwork().\
                internal_average_path_length([1,2,4], None))
        1.0

        :arg [int] node_list: list of node indices describing the subnetwork
        :arg str link_attribute: Optional name of the link attribute to be used
            as the links' length. If None, links have length 1. (Default: None)
        :return float: the internal average path length.
        """
        path_lengths = self.internal_path_lengths(node_list, link_attribute)
        return self._calculate_general_average_path_length(path_lengths,
                                                           internal=True)

    def average_cross_closeness(self, node_list1, node_list2,
                                link_attribute=None):
        """
        Return the average cross closeness.

        **Example:**

        >>> r(InteractingNetworks.SmallTestNetwork().\
                average_cross_closeness([0,5], [1,2,3,4]))
        1.7143

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :arg str link_attribute: Optional name of the link attribute to be used
            as the links' length. If None, links have length 1. (Default: None)
        :return float: the average cross closeness.
        """
        return np.mean(self.cross_closeness(node_list1, node_list2,
                                            link_attribute))

    def global_efficiency(self, node_list1, node_list2, link_attribute=None):
        """
        Return the global efficiency.

        **Example:**

        >>> r(InteractingNetworks.SmallTestNetwork().\
                global_efficiency([0,5], [1,2,3,4]))
        1.7143

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :arg str link_attribute: Optional name of the link attribute to be used
            as the links' length. If None, links have length 1. (Default: None)
        :return float: the global efficiency.
        """
        local_efficiency = self.local_efficiency(node_list1, node_list2,
                                                 link_attribute)
        return 1/np.mean(local_efficiency)

    #
    #  Define local measures for interacting networks
    #

    def cross_degree(self, node_list1, node_list2, link_attribute=None):
        """
        Return the cross degree sequence for one subnetwork with respect to a
        second subnetwork.

        Gives the number of links from a specific node in the first subnetwork
        projecting to the second subnetwork. If a link attribute is specified,
        return the associated strength

        **Examples:**

        >>> InteractingNetworks.SmallTestNetwork().\
                cross_degree([0,3,5], [1,2,4])
        array([1, 1, 0])
        >>> InteractingNetworks.SmallTestNetwork().\
                cross_degree([1,2,4], [0,3,5])
        array([1, 0, 1])
        >>> InteractingNetworks.SmallTestNetwork().\
                cross_degree([1,2,3,4], [0,5])
        array([0, 0, 1, 1])

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :arg str key: link attribute key (optional)
        :rtype: 1D array [node index]
        :return: the cross degree sequence.
        """
        if self.directed:
            return (self.cross_indegree(node_list1, node_list2,
                                        link_attribute)
                    + self.cross_outdegree(node_list1, node_list2,
                                           link_attribute))
        else:
            return self.cross_outdegree(node_list1, node_list2,
                                        link_attribute)

    def cross_indegree(self, node_list1, node_list2, link_attribute=None):
        """
        Return the cross indegree sequence for the first given subnetwork with
        respect to the second given subnetwork

        Gives the number of links from nodes in subnetwork two to a specific
        node from subnetwork one. If a link attribute is specified,
        return the associated in strength.


        **Example:**

        >>> net = InteractingNetworks.SmallDirectedTestNetwork()
        >>> net.cross_indegree([1, 2], [0, 3, 4])
        array([2, 1])

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :arg str key: link attribute key (optional)
        :rtype: 1D array [node index]
        :return: the cross in degree sequence.
        """
        if link_attribute is None:
            return np.sum(self.cross_adjacency(node_list2, node_list1), axis=0)
        else:
            return np.sum(self.cross_link_attribute(link_attribute, node_list2,
                                                    node_list1), axis=0)

    def cross_outdegree(self, node_list1, node_list2, link_attribute=None):
        """
        Return the cross outdegree sequence for the first given subnetwork with
        respect to the second given subnetwork

        Gives the number of links from a specific node in subnetwork one to
        nodes in subnetwork two. If a link attribute is specified,
        return the associated out strength.

        **Example:**

        >>> net = InteractingNetworks.SmallDirectedTestNetwork()
        >>> net.cross_outdegree([1, 2], [0, 3, 4])
        array([1, 0])

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :arg str key: link attribute key (optional)
        :rtype: 1D array [node index]
        :return: the cross out degree sequence.
        """
        if link_attribute is None:
            return np.sum(self.cross_adjacency(node_list1, node_list2), axis=1)
        else:
            return np.sum(self.cross_link_attribute(link_attribute, node_list1,
                                                    node_list2), axis=1)

    def internal_degree(self, node_list, link_attribute=None):
        """
        Return the internal degree sequence of one induced subnetwork.

        Gives the number of links from a specific node to other nodes within
        the same induced subnetwork. If a link attribute is specified,
        return the associated strength.

        **Examples:**

        >>> InteractingNetworks.SmallTestNetwork().internal_degree([0,3,5])
        array([2, 1, 1])
        >>> InteractingNetworks.SmallTestNetwork().internal_degree([1,2,4])
        array([2, 2, 2])

        :arg [int] node_list: list of node indices describing the subnetwork
        :arg str key: link attribute key (optional)
        :rtype: 1D array [node index]
        :return: the internal degree sequence.
        """
        if self.directed:
            return (self.internal_indegree(node_list, link_attribute)
                    + self.internal_outdegree(node_list, link_attribute))
        else:
            return self.internal_outdegree(node_list, link_attribute)

    def internal_indegree(self, node_list, link_attribute=None):
        """
        Return the internal indegree sequence of one induced subnetwork.

        Gives the number of links from other nodes to a specific node within
        the same induced subnetwork. If a link attribute is specified,
        return the associated in strength.

        **Example:**

        >>> net = InteractingNetworks.SmallDirectedTestNetwork()
        >>> net.internal_indegree([0, 1, 3])
        array([0, 2, 1])

        :arg [int] node_list: list of node indices describing the subnetwork
        :arg str key: link attribute key (optional)
        :rtype: 1D array [node index]
        :return: the internal in degree sequence.
        """
        if link_attribute is None:
            return np.sum(self.internal_adjacency(node_list), axis=0)
        else:
            return np.sum(self.internal_link_attribute(link_attribute,
                                                       node_list), axis=0)

    def internal_outdegree(self, node_list, link_attribute=None):
        """
        Return the internal outdegree sequence of one induced subnetwork.

        Gives the number of links from a specific node to other nodes within
        the same induced subnetwork. If a link attribute is specified,
        return the associated out strength.

        **Example:**

        >>> net = InteractingNetworks.SmallDirectedTestNetwork()
        >>> net.internal_outdegree([0, 1, 3])
        array([2, 0, 1])

        :arg [int] node_list: list of node indices describing the subnetwork
        :arg str key: link attribute key (optional)
        :rtype: 1D array [node index]
        :return: the internal out degree sequence.
        """
        if link_attribute is None:
            return np.sum(self.internal_adjacency(node_list), axis=1)
        else:
            return np.sum(self.internal_link_attribute(link_attribute,
                                                       node_list), axis=1)

    def cross_local_clustering(self, node_list1, node_list2):
        """
        Return local cross clustering for a pair of subnetworks.

        The local cross clustering coefficient C_v gives the probability, that
        two randomly drawn neighbors in subnetwork 1 of node v in subnetwork 1
        are also neighbors. It counts triangles having one vertex in
        subnetwork 1 and two vertices in subnetwork 2.

        **Examples:**

        >>> InteractingNetworks.SmallTestNetwork().\
                cross_local_clustering([0,3,5], [1,2,4])
        array([ 0.,  0.,  0.])
        >>> InteractingNetworks.SmallTestNetwork().\
                cross_local_clustering([2], [1,3,4])
        array([ 1.])
        >>> InteractingNetworks.SmallTestNetwork().\
                cross_local_clustering([3,4], [1,2])
        array([ 0.,  1.])

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :rtype: 1D array [node index]
        :return: the cross local clustering coefficient.
        """
        nodes1 = np.array(node_list1, dtype=NODE)
        nodes2 = np.array(node_list2, dtype=NODE)
        #  Get cross degree sequence
        cross_degree = InteractingNetworks.cross_degree(self, nodes1, nodes2)
        #  Prepare normalization factor
        norm = cross_degree * (cross_degree - 1) / 2.
        #  Initialize
        cross_clustering = np.zeros_like(nodes1, dtype=DFIELD)

        _cross_local_clustering(to_cy(self.adjacency, ADJ),
                                to_cy(norm, DFIELD),
                                nodes1, nodes2, cross_clustering)
        return cross_clustering

    def cross_local_clustering_sparse(self, node_list1, node_list2):
        """
        Return local cross clustering for a pair of subnetworks.

        The local cross clustering coefficient C_v gives the probability, that
        two randomly drawn neighbors in subnetwork 1 of node v in subnetwork 1
        are also neighbors. It counts triangles having one vertex in
        subnetwork 1 and two vertices in subnetwork 2.

        Examples:

        >>> InteractingNetworks.SmallTestNetwork().\
                cross_local_clustering_sparse([0,3,5], [1,2,4])
        array([ 0.,  0.,  0.])

        >>> InteractingNetworks.SmallTestNetwork().\
                cross_local_clustering_sparse([2], [1,3,4])
        array([ 1.])

        >>> InteractingNetworks.SmallTestNetwork().\
                cross_local_clustering_sparse([3,4], [1,2])
        array([ 0.,  1.])

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :rtype: 1D array [node index]
        :return: the cross local clustering coefficient.
        """
        #  Get cross degree sequence
        cross_degree = self.cross_degree(node_list1, node_list2)
        #  Get full adjacency matrix
        A = self.sp_A[node_list1+node_list2, :][:, node_list1+node_list2]
        #  Get layer sizes
        N1, N2 = len(node_list1), len(node_list2)
        #  Convert node lists to Numpy arrays
        nodes1, nodes2 = np.array(node_list1), np.array(node_list2)

        #  Initialize
        cross_clustering = np.zeros(N1)
        #  Prepare normalization factor
        norm = cross_degree * (cross_degree - 1) / 2

        # Calculate cross clustering from subnetwork 1 to subnetwork 2
        counter = 0
        for node1 in range(N1):
            if not norm[node1] == 0:
                # Reset counter
                counter = 0
                #  Loop over unique pairs of nodes in subnetwork 2
                for node2 in range(N1, N1+N2):
                    for node3 in range(N1, node2):
                        if (A[node1, node2] == 1 and A[node2, node3] == 1
                                and A[node3, node1] == 1):
                            counter += 1
                cross_clustering[node1] = counter / norm[node1]

        return cross_clustering

    def _calculate_general_closeness(self, path_lengths, internal=True):
        """
        Calculate general closeness sequence for interacting networks.

        :type path_lengths: 2D array [node,node] of floats
        :arg  path_lengths: Path lengths to use
        :arg bool internal: Indicates, whether internal or cross closeness
            shall be calculated.
        :rtype:  1D array [index]
        :return: the general closeness sequence.
        """
        #  Get shape of path lengths array
        (N, M) = path_lengths.shape

        #  Set total number of nodes to be considered for calculation
        if internal:
            n_nodes = N
            norm = M - 1
        else:
            n_nodes = self.N  # All nodes of the whole network here!
            norm = M

        #  Closeness has the length of the first dimension of path lengths
        closeness = np.zeros(N)

        #  Identify unconnected pairs and save in binary array isinf
        unconnected_pairs = np.isinf(path_lengths)
        #  Set infinite entries corresponding to unconnected pairs to maximum
        #  possible path length.
        path_lengths[unconnected_pairs] = n_nodes - 1

        #  Some nodes have a distance of zero to all their
        #  neighbors. These nodes get zero closeness centrality.
        path_length_sum = path_lengths.sum(axis=1)
        #  M entries have been summed over, so we also have to normalize by M
        closeness[path_length_sum != 0] = \
            norm / path_length_sum[path_length_sum != 0]

        #  Reverse changes to path_lengths
        path_lengths[unconnected_pairs] = np.inf

        return closeness

    def cross_closeness(self, node_list1, node_list2, link_attribute=None):
        """
        Return cross closeness sequence for a pair of induced subnetworks.

        Gives the inverse average geodesic distance from a node in subnetwork 1
        to all nodes in subnetwork 2.

        **Examples:**

        >>> InteractingNetworks.SmallTestNetwork().\
                cross_closeness([0,3,5], [1,2,4], None)
        array([ 0.6  ,  0.6  ,  0.375])
        >>> InteractingNetworks.SmallTestNetwork().\
                cross_closeness([0,5], [1,2,3,4], None)
        array([ 0.66666667,  0.4       ])

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :arg str link_attribute: Optional name of the link attribute to be used
            as the links' length. If None, links have length 1. (Default: None)
        :rtype: 1D arrays [index]
        :return: the cross closeness sequence.
        """
        path_lengths = InteractingNetworks.cross_path_lengths(
            self, node_list1, node_list2, link_attribute)

        return self._calculate_general_closeness(path_lengths, internal=False)

    def internal_closeness(self, node_list, link_attribute=None):
        """
        Return internal closeness sequence for an induced subnetwork.

        Gives the inverse average geodesic distance from a node to all other
        nodes in the same induced subnetwork.

        However, the included paths will generally contain nodes from the whole
        network. To avoid this, do the following:

        >>> r(InteractingNetworks.SmallTestNetwork().\
                subnetwork([0,3,5]).closeness(None))
        array([ 1. , 0.6667, 0.6667])

        **Examples:**

        >>> InteractingNetworks.SmallTestNetwork().\
                internal_closeness([0,3,5], None)
        array([ 1.        ,  0.66666667,  0.66666667])
        >>> InteractingNetworks.SmallTestNetwork().\
                internal_closeness([1,2,4], None)
        array([ 1.,  1.,  1.])

        :arg [int] node_list: list of node indices describing the subnetwork
        :arg str link_attribute: Optional name of the link attribute to be used
            as the links' length. If None, links have length 1. (Default: None)
        :rtype: 1D array [index]
        :return: the internal closeness sequence.
        """
        path_lengths = self.internal_path_lengths(node_list, link_attribute)
        return self._calculate_general_closeness(path_lengths, internal=True)

    def cross_betweenness(self, node_list1, node_list2):
        """
        Return the cross betweenness sequence for the whole network with
        respect to a pair of subnetworks.

        Gives the normalized number of shortest paths only between nodes from
        **two** subnetworks, in which a node :math:`i` is contained. This is
        equivalent to the inter-regional / inter-group betweenness with respect
        to subnetwork 1 and subnetwork 2.

        **Examples:**

        >>> InteractingNetworks.SmallTestNetwork().\
                cross_betweenness([2], [3,5])
        array([ 1.,  1.,  0.,  0.,  1.,  0.])
        >>> InteractingNetworks.SmallTestNetwork().\
                cross_betweenness(range(0,6), range(0,6))
        array([ 9.,  3.,  0.,  2.,  6.,  0.])

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :rtype: 1D arrays [node index]
        :return: the cross betweenness sequence for the whole network with
            respect to two subnetworks.
        """
        return self.interregional_betweenness(sources=node_list1,
                                              targets=node_list2)

    def internal_betweenness(self, node_list):
        """
        Return the internal betweenness sequence for an induced subnetwork.

        Gives the normalized number of shortest paths only between nodes from
        subnetwork 1, in which a node :math:`i` from the whole network is
        contained.  This is equivalent to the inter-regional / inter-group
        betweenness with respect to subnetwork 1 and subnetwork 1.

        **Examples:**

        >>> InteractingNetworks.SmallTestNetwork().\
                internal_betweenness(range(0,6))
        array([ 9.,  3.,  0.,  2.,  6.,  0.])

        :arg [int] node_list: list of node indices describing the subnetwork
        :rtype: 1D array [node index]
        :return: the internal betweenness sequence for layer 1.
        """
        return self.interregional_betweenness(sources=node_list,
                                              targets=node_list)

    def local_efficiency(self, node_list1, node_list2, link_attribute=None):
        """
        Return the local efficiency sequence for an induced subnetwork.

        **Example:**

        >>> InteractingNetworks.SmallTestNetwork().\
                local_efficiency([0,5], [1,2,3,4])
        array([0.75      , 0.41666667])

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :arg str link_attribute: Optional name of the link attribute to be used
            as the links' length. If None, links have length 1. (Default: None)
        :rtype: 1D arrays [index]
        :return: the local efficiency sequence.
        """
        path_lengths = self.cross_path_lengths(node_list1, node_list2,
                                               link_attribute)
        return np.mean(1/path_lengths, axis=1)

    def nsi_cross_degree(self, node_list1, node_list2):
        """
        Return the n.s.i. cross-degree for a pair of induced subnetworks.

        Gives an estimation about the quota of the whole domain of interest of
        the subnetwork 2 any node in the subnetwork 1 is connected to.

        **Examples:**

        >>> InteractingNetworks.SmallTestNetwork().\
                nsi_cross_degree([0,1,2],[3,4,5])
        array([ 4.2,  2.6,  1.4])
        >>> InteractingNetworks.SmallTestNetwork().\
                nsi_cross_degree([0,2,5],[1,4])
        array([ 1.4,  2.2,  0. ])

        :arg [int] node_list1: list of node indices describing the subnetwork 1
        :arg [int] node_list2: list of node indices describing the subnetwork 2
        :rtype: 1D array [node index]
        :return: the n.s.i. cross-degree for layer 1.
        """
        cross_A = (self.adjacency
                   + np.eye(self.N))[node_list1, :][:, node_list2]
        return (cross_A * self.node_weights[node_list2]).sum(axis=1)

    def nsi_cross_mean_degree(self, node_list1, node_list2):
        """
        Return the n.s.i. cross-mean degree for a pair of induced subnetworks.

        **Examples:**

        >>> InteractingNetworks.SmallTestNetwork().\
                nsi_cross_mean_degree([0,1,2],[3,4,5])
        2.5
        >>> r(InteractingNetworks.SmallTestNetwork().\
                nsi_cross_mean_degree([0,2,5],[1,4]))
        0.95

        :arg [int] node_list1: list of node indices describing the subnetwork 1
        :arg [int] node_list2: list of node indices describing the subnetwork 2
        :return float: the n.s.i. cross-mean degree for layer 1.
        """
        nsi_cross = self.nsi_cross_degree(node_list1, node_list2)
        node_weights = self.node_weights[node_list1]
        W_i = sum(node_weights)
        return sum(nsi_cross * node_weights) / W_i

    def nsi_internal_degree(self, node_list):
        """
        Return the n.s.i. internal degree sequence of one induced subnetwork.

        **Examples:**

        >>> InteractingNetworks.SmallTestNetwork().\
                nsi_internal_degree([0,3,5])
        array([ 3.4,  1.8,  2.2])
        >>> InteractingNetworks.SmallTestNetwork().\
                nsi_internal_degree([0,1,3,5])
        array([ 3.4,  2. ,  2.6,  2.2])

        :arg [int] node_list: list of node indices describing the subnetwork
        :rtype: 1D array [node index]
        :return: the n.s.i. internal degree sequence
        """
        return self.nsi_cross_degree(node_list, node_list)

    def nsi_cross_local_clustering(self, node_list1, node_list2):
        """
        Return the n.s.i. cross-local clustering coefficient for a pair of
        induced subnetworks.

        **Examples:**

        >>> InteractingNetworks.SmallTestNetwork().\
                nsi_cross_local_clustering([0,1,2],[3,4,5])
        array([ 0.33786848,  0.50295858,  1.  ])
        >>> InteractingNetworks.SmallTestNetwork().\
                nsi_cross_local_clustering([0,2,5],[1,4])
        array([ 1.,  1.,  0.])

        :arg [int] node_list1: list of node indices describing the subnetwork 1
        :arg [int] node_list2: list of node indices describing the subnetwork 2
        :rtype: 1D array [node index]
        :return: the n.s.i. cross-local clustering coefficient for layer 1.
        """
        nodes1 = np.array(node_list1, dtype=NODE)
        nodes2 = np.array(node_list2, dtype=NODE)
        nsi_cc = np.zeros(nodes1.shape, dtype=DFIELD)
        _nsi_cross_local_clustering(
            to_cy(self.adjacency + np.eye(self.N, dtype=ADJ), ADJ),
            nsi_cc, nodes1, nodes2, to_cy(self.node_weights, DWEIGHT))

        norm = self.nsi_cross_degree(nodes1, nodes2) ** 2
        nsi_cc[norm != 0] = nsi_cc[norm != 0] / norm[norm != 0]
        nsi_cc[norm == 0] = 0
        return nsi_cc

    def nsi_cross_closeness_centrality(self, node_list1, node_list2):
        """
        Return the n.s.i. cross-closeness centrality for a pair of induced
        subnetworks.

        **Examples:**

        >>> InteractingNetworks.SmallTestNetwork().\
                nsi_cross_closeness_centrality([0,1,2],[3,4,5])
        array([ 1.        ,  0.56756757,  0.48837209])
        >>> InteractingNetworks.SmallTestNetwork().\
                nsi_cross_closeness_centrality([0,2,5],[1,4])
        array([ 0.73333333,  1.        ,  0.42307692])

        :arg [int] node_list1: list of node indices describing the subnetwork 1
        :arg [int] node_list2: list of node indices describing the subnetwork 2
        :rtype: 1D array [node index]
        :return: the n.s.i. cross-closeness for layer 1.
        """
        shortest_paths = self.path_lengths()
        node_weights = self.node_weights

        nsi_shortest_paths = shortest_paths + np.eye(len(shortest_paths))
        nsi_shortest_paths[np.isinf(nsi_shortest_paths)] = self.N - 1

        nsi_cross_paths = nsi_shortest_paths[node_list1, :][:, node_list2]
        W = sum(node_weights[node_list2])
        return W / np.dot(nsi_cross_paths, node_weights[node_list2])

    def nsi_internal_closeness_centrality(self, node_list):
        """
        Return the n.s.i. internal closeness sequence of one induced
        subnetwork.

        **Examples:**

        >>> InteractingNetworks.SmallTestNetwork().\
                nsi_internal_closeness_centrality([0,3,5])
        array([ 1.        ,  0.68      ,  0.73913043])
        >>> InteractingNetworks.SmallTestNetwork().\
                nsi_internal_closeness_centrality([0,1,3,5])
        array([ 0.84      ,  0.525     ,  0.72413793,  0.6       ])

        :arg [int] node_list: list of node indices describing the subnetwork
        :rtype: 1D array [node index]
        :return: the n.s.i. internal closeness sequence
        """
        return self.nsi_cross_closeness_centrality(node_list, node_list)

    def nsi_cross_global_clustering(self, node_list1, node_list2):
        """
        Return the n.s.i. cross-global clustering coefficient for an induced
        subnetwork 1 with regard to a second induced subnetwork 2.

        **Examples:**

        >>> r(InteractingNetworks.SmallTestNetwork().\
                nsi_cross_global_clustering([0,1,2],[3,4,5]))
        0.6688

        :arg [int] node_list1: list of node indices describing the subnetwork 1
        :arg [int] node_list2: list of node indices describing the subnetwork 2
        :return float: the n.s.i. cross-global clustering coefficient for the
            subnetwork 1 with regard to subnetwork 2.
        """
        nsi_cc = self.nsi_cross_local_clustering(node_list1, node_list2)
        node_weights = self.node_weights[node_list1]
        return sum(node_weights * nsi_cc) / sum(node_weights)

    def nsi_internal_local_clustering(self, node_list):

        """
        Return the n.s.i. internal cross-local clustering coefficient for an
        induced subnetwork.

        **Examples:**

        >>> InteractingNetworks.SmallTestNetwork().\
                nsi_internal_local_clustering([1,2,3,5])
        array([ 0.73333333,  1.        ,  1.        ,  1.        ])
        >>> InteractingNetworks.SmallTestNetwork().\
                nsi_internal_local_clustering([0,2,4])
        array([ 1.        ,  1.        ,  0.86666667])

        :arg [int] node_list: list of node indices describing the subnetwork
        :rtype: 1D numpy array [node_index]
        :return: the n.s.i. internal-local clustering coefficient for all nodes
            within the induced subnetwork
        """
        return self.nsi_cross_local_clustering(node_list, node_list)

    def nsi_cross_betweenness(self, node_list1, node_list2):
        """
        Return the n.s.i. cross betweenness sequence for the whole network with
        respect to a pair of subnetworks.

        **Examples:**

        >>> r(InteractingNetworks.SmallTestNetwork().\
                nsi_cross_betweenness([0,4,5],[1,3]))
        array([ 6.5333,  1.2   ,  0.    ,
                0.6769,  0.6769,  0.    ])
        >>> r(InteractingNetworks.SmallTestNetwork().\
                nsi_cross_betweenness([0,1],[2,3,4,5]))
        array([ 2.1333,  0.    ,  0.    ,
                0.4923,  0.9209,  0.    ])

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :rtype: 1D arrays [node index]
        :return: the n.s.i. cross betweenness sequence for the whole network
            with respect to two subnetworks.
        """
        return self.nsi_interregional_betweenness(sources=node_list1,
                                                  targets=node_list2)

    def nsi_cross_edge_density(self, node_list1, node_list2):
        """
        Return the n.s.i. density of edges between two subnetworks.

        **Examples:**

        >>> r(InteractingNetworks.SmallTestNetwork().\
                nsi_cross_edge_density([1,2,3],[0,5]))
        0.1091
        >>> r(InteractingNetworks.SmallTestNetwork().\
                nsi_cross_edge_density([0],[1,4,5]))
        0.7895

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :return float: the n.s.i. cross density of edges between two
            subnetworks 1 and 2.
        """
        W_j = sum(self.node_weights[node_list2])
        return self.nsi_cross_mean_degree(node_list1, node_list2) / W_j

    def nsi_cross_transitivity(self, node_list1, node_list2):
        """
        Return n.s.i. cross-transitivity for a pair of subnetworks.

        **Examples:**

        >>> r(InteractingNetworks.SmallTestNetwork().\
                nsi_cross_transitivity([1,2],[0,3,4,5]))
        0.6352
        >>> InteractingNetworks.SmallTestNetwork().\
                nsi_cross_transitivity([0,2,3],[1])
        1.0


        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :return float: the n.s.i. cross transitivity for a pair of subnetworks
            1 and 2.
        """
        return _nsi_cross_transitivity(
            to_cy(self.adjacency + np.eye(self.N, dtype=ADJ), ADJ),
            np.array(node_list1, dtype=NODE),
            np.array(node_list2, dtype=NODE),
            to_cy(self.node_weights, DWEIGHT))

    def nsi_cross_average_path_length(self, node_list1, node_list2):
        """
        Return n.s.i. cross average path length between two induced
        subnetworks.

        **Examples:**

        >>> net = InteractingNetworks.SmallTestNetwork()
        >>> r(net.nsi_cross_average_path_length([0,5],[1,2,4]))
        3.3306
        >>> r(net.nsi_cross_average_path_length([1,3,4,5],[2]))
        0.376

        :arg [int] node_list1: list of node indices describing the first
            subnetwork
        :arg [int] node_list2: list of node indices describing the second
            subnetwork
        :return float: the n.s.i. cross-average path length between a pair of
            subnetworks.
        """
        shortest_paths = self.path_lengths()
        nsi_shortest_paths = shortest_paths + np.eye(len(shortest_paths))
        node_weights = self.node_weights

        Wi = sum(node_weights[node_list1])
        Wj = sum(node_weights[node_list1])

        w_v = np.zeros([len(node_list2), len(node_list1)])
        w_v[:] = node_weights[node_list1]
        w_q = np.zeros([len(node_list1), len(node_list2)])
        w_q[:] = node_weights[node_list2]

        Wij = w_v.transpose() + w_q
        nsi_cross_paths = nsi_shortest_paths[node_list1, :][:, node_list2]
        Wij = Wij[np.isinf(nsi_cross_paths)].sum()
        nsi_shortest_paths[np.isinf(nsi_shortest_paths)] = self.N - 1
        nsi_cross_paths = nsi_shortest_paths[node_list1, :][:, node_list2]

        Lij = np.sum(nsi_cross_paths*node_weights[node_list2], axis=1)
        Lij = np.sum(Lij * node_weights[node_list1], axis=0)
        return Lij / (Wi*Wj - Wij)



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Generic consistency checks.
"""

from pyunicorn import __version__, core, climate, timeseries, funcnet


def test_version():
    """
    Valid version information.
    """
    assert isinstance(__version__, str)


def test_init_str():
    """
    Reasonable __init__ and __str__ for most classes.
    """
    for i in simple_instances():
        assert str(i).startswith(i.__class__.__name__)


def simple_instances():
    """
    Minimally initialized instances for most classes.
    """
    cd = climate.ClimateData.SmallTestData()
    ca = funcnet.CouplingAnalysis.test_data()[:50]
    ts = ca[:, 0]
    t1, t2, t3 = [{'threshold': t} for t in [0.2, (0.2, 0.2), (0.2, 0.2, 0.2)]]
    # es = 1-(np.random.rand(100, 10) > 0.4).astype(int)
    # ec = climate.EventSynchronizationClimateNetwork.SmallTestData()
    return [
        core.Network.SmallTestNetwork(),
        core.Grid.SmallTestGrid(),
        core.SpatialNetwork.SmallTestNetwork(),
        core.GeoGrid.SmallTestGrid(),
        core.GeoNetwork.SmallTestNetwork(),
        core.InteractingNetworks.SmallTestNetwork(),
        core.ResNetwork.SmallTestNetwork(),
        core.NetCDFDictionary(),
        cd,
        climate.ClimateNetwork.SmallTestNetwork(),
        climate.HavlinClimateNetwork(cd, 0, **t1),
        climate.HilbertClimateNetwork(cd, **t1),
        climate.TsonisClimateNetwork.SmallTestNetwork(),
        climate.PartialCorrelationClimateNetwork(cd, winter_only=False, **t1),
        climate.RainfallClimateNetwork(cd, **t1),
        climate.SpearmanClimateNetwork(cd, winter_only=False, **t1),
        climate.MutualInfoClimateNetwork(cd, winter_only=False, **t1),
        climate.CoupledTsonisClimateNetwork(cd, cd, **t1),
        # climate.EventSynchronizationClimateNetwork(
        #     ec, 0.8, 16, eventsynctype="directedES"),
        timeseries.RecurrencePlot(ts, **t1),
        timeseries.RecurrenceNetwork(ts, **t1),
        timeseries.JointRecurrencePlot(ts, ts, **t2),
        timeseries.JointRecurrenceNetwork(ts, ts, **t2),
        timeseries.CrossRecurrencePlot(ts, ts, **t1),
        timeseries.InterSystemRecurrenceNetwork(ts, ts, **t3),
        timeseries.Surrogates.SmallTestData(),
        timeseries.VisibilityGraph(ts),
        funcnet.CouplingAnalysis(ca),
        funcnet.CouplingAnalysisPurePython(ca),
        # funcnet.EventSynchronization(es, 16)
    ]



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Simple tests for the RecurrencePlot class.
"""

from itertools import chain, product

import pytest
import numpy as np

from pyunicorn.core import Data
from pyunicorn.core._ext.types import NODE, ADJ, DFIELD
from pyunicorn.timeseries import RecurrencePlot


# -----------------------------------------------------------------------------
# test RecurrencePlot instantiation
# -----------------------------------------------------------------------------

# test non-default metrics

def test_RP_euclidean():
    x = Data.SmallTestData().observable()
    RP = RecurrencePlot(x[:, :-3], threshold=1.2, metric='euclidean')
    res = RP.recurrence_matrix()
    assert res.dtype == ADJ
    exp = np.array([
        [1, 1, 1, 0, 0, 0, 0, 0, 0, 0],
        [1, 1, 1, 1, 1, 0, 0, 0, 0, 0],
        [1, 1, 1, 1, 1, 1, 0, 0, 0, 0],
        [0, 1, 1, 1, 1, 1, 1, 1, 0, 0],
        [0, 1, 1, 1, 1, 1, 1, 1, 0, 0],
        [0, 0, 1, 1, 1, 1, 1, 1, 1, 0],
        [0, 0, 0, 1, 1, 1, 1, 1, 1, 1],
        [0, 0, 0, 1, 1, 1, 1, 1, 1, 1],
        [0, 0, 0, 0, 0, 1, 1, 1, 1, 1],
        [0, 0, 0, 0, 0, 0, 1, 1, 1, 1]
    ])
    assert np.array_equal(res, exp)


def test_RP_manhattan():
    x = Data.SmallTestData().observable()
    RP = RecurrencePlot(x, threshold=3.5, metric='manhattan')
    res = RP.recurrence_matrix()
    assert res.dtype == ADJ
    exp = np.array([
        [1, 1, 1, 0, 0, 0, 0, 0, 0, 0],
        [1, 1, 1, 1, 0, 0, 0, 0, 0, 0],
        [1, 1, 1, 1, 1, 0, 0, 0, 0, 0],
        [0, 1, 1, 1, 1, 1, 1, 0, 0, 0],
        [0, 0, 1, 1, 1, 1, 1, 1, 0, 0],
        [0, 0, 0, 1, 1, 1, 1, 1, 0, 0],
        [0, 0, 0, 1, 1, 1, 1, 1, 1, 0],
        [0, 0, 0, 0, 1, 1, 1, 1, 1, 1],
        [0, 0, 0, 0, 0, 0, 1, 1, 1, 1],
        [0, 0, 0, 0, 0, 0, 0, 1, 1, 1]
    ])
    assert np.array_equal(res, exp)


# test thresholding variations

def test_RP_threshold_std():
    x = Data.SmallTestData().observable()
    RP = RecurrencePlot(x, threshold_std=.8)
    res = RP.recurrence_matrix()
    assert res.dtype == ADJ
    exp = np.array([
        [1, 1, 0, 0, 0, 0, 0, 0, 0, 0],
        [1, 1, 1, 1, 0, 0, 0, 0, 0, 0],
        [0, 1, 1, 1, 1, 0, 0, 0, 0, 0],
        [0, 1, 1, 1, 1, 0, 0, 0, 0, 0],
        [0, 0, 1, 1, 1, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 1, 1, 1, 0, 0, 0],
        [0, 0, 0, 0, 0, 1, 1, 1, 1, 0],
        [0, 0, 0, 0, 0, 0, 1, 1, 1, 1],
        [0, 0, 0, 0, 0, 0, 1, 1, 1, 1],
        [0, 0, 0, 0, 0, 0, 0, 1, 1, 1]
    ])
    assert np.array_equal(res, exp)


def test_RP_recurrence_rate():
    x = Data.SmallTestData().observable()
    RP = RecurrencePlot(x, recurrence_rate=.4)
    res = RP.recurrence_matrix()
    assert res.dtype == ADJ
    exp = np.array([
        [1, 1, 0, 0, 0, 0, 0, 0, 0, 0],
        [1, 1, 1, 1, 0, 0, 0, 0, 0, 0],
        [0, 1, 1, 1, 1, 0, 0, 0, 0, 0],
        [0, 1, 1, 1, 1, 1, 0, 0, 0, 0],
        [0, 0, 1, 1, 1, 1, 0, 0, 0, 0],
        [0, 0, 0, 1, 1, 1, 1, 0, 0, 0],
        [0, 0, 0, 0, 0, 1, 1, 1, 1, 0],
        [0, 0, 0, 0, 0, 0, 1, 1, 1, 1],
        [0, 0, 0, 0, 0, 0, 1, 1, 1, 1],
        [0, 0, 0, 0, 0, 0, 0, 1, 1, 1]
    ])
    assert np.array_equal(res, exp)


def test_RP_local_recurrence_rate():
    x = Data.SmallTestData().observable()
    RP = RecurrencePlot(x, local_recurrence_rate=.6)
    res = RP.recurrence_matrix()
    assert res.dtype == ADJ
    exp = np.array([
        [1, 1, 1, 1, 1, 0, 0, 0, 0, 0],
        [1, 1, 1, 1, 1, 0, 0, 0, 0, 0],
        [1, 1, 1, 1, 1, 0, 0, 0, 0, 0],
        [0, 1, 1, 1, 1, 1, 0, 0, 0, 0],
        [0, 0, 1, 1, 1, 1, 1, 0, 0, 0],
        [0, 0, 0, 1, 1, 1, 1, 1, 0, 0],
        [0, 0, 0, 0, 1, 1, 1, 1, 1, 0],
        [0, 0, 0, 0, 0, 1, 1, 1, 1, 1],
        [0, 0, 0, 0, 0, 1, 1, 1, 1, 1],
        [0, 0, 0, 0, 0, 1, 1, 1, 1, 1]
    ])
    assert np.array_equal(res, exp)


# -----------------------------------------------------------------------------
# prepare fixtures
# -----------------------------------------------------------------------------


@pytest.fixture(scope="module", name="recurrence_crit", ids=str,
                params=list(chain(product(np.arange(0, 1.7, .8), [None]),
                                  product([None], np.arange(0, 1.1, .4)))))
def recurrence_crit_fixture(request):
    threshold, rate = request.param
    assert np.sum([threshold is None, rate is None]) == 1
    return request.param


@pytest.fixture(scope="module", name="small_RP")
def small_RP_fixture(metric, recurrence_crit):
    """
    RP fixture, parametrized to cover various settings.
    """
    x = Data.SmallTestData().observable()
    threshold, rate = recurrence_crit
    return RecurrencePlot(
        x, threshold=threshold, recurrence_rate=rate, metric=metric)


@pytest.fixture(scope="module", name="small_RP_basic", params=[False, True])
def small_RP_basic_fixture(request):
    """
    RP fixture with single basic setting to test numerical results.
    """
    sparse = request.param
    x = Data.SmallTestData().observable()
    RP = RecurrencePlot(x, threshold=.8, metric='supremum', sparse_rqa=sparse)
    if not sparse:
        res = RP.recurrence_matrix()
        assert res.dtype == ADJ
        exp = np.array([
            [1, 1, 1, 0, 0, 0, 0, 0, 0, 0],
            [1, 1, 1, 1, 1, 0, 0, 0, 0, 0],
            [1, 1, 1, 1, 1, 0, 0, 0, 0, 0],
            [0, 1, 1, 1, 1, 1, 0, 0, 0, 0],
            [0, 1, 1, 1, 1, 1, 1, 0, 0, 0],
            [0, 0, 0, 1, 1, 1, 1, 1, 0, 0],
            [0, 0, 0, 0, 1, 1, 1, 1, 1, 1],
            [0, 0, 0, 0, 0, 1, 1, 1, 1, 1],
            [0, 0, 0, 0, 0, 0, 1, 1, 1, 1],
            [0, 0, 0, 0, 0, 0, 1, 1, 1, 1]
        ])
        assert np.array_equal(res, exp)
    return RP


# -----------------------------------------------------------------------------
# test RecurrencePlot RQA
# -----------------------------------------------------------------------------


@pytest.mark.parametrize("measure", ["diag", "vert", "white_vert"])
def test_line_dist(measure: str, small_RP):
    res = getattr(small_RP, f"{measure}line_dist")()
    assert res.dtype == NODE
    assert res.shape[0] == small_RP.N
    assert (0 <= res).all() and (res <= small_RP.N).all()


@pytest.mark.parametrize(
    "measure, exp",
    [("diag", [4, 0, 0, 0, 0, 0, 0, 2, 2, 0]),
     ("vert", [0, 0, 1, 2, 5, 2, 0, 0, 0, 0]),
     ("white_vert", [2, 1, 2, 2, 3, 2, 1, 0, 0, 0])])
def test_line_dist_numeric(measure: str, small_RP_basic, exp):
    if small_RP_basic.sparse_rqa and measure == "white_vert":
        return
    res = getattr(small_RP_basic, f"{measure}line_dist")()
    assert res.dtype == NODE
    assert res.shape[0] == small_RP_basic.N
    assert np.array_equal(res, exp)


@pytest.mark.parametrize("sparse", [False, True])
def test_line_dist_edgecases(sparse):
    x = Data.SmallTestData().observable()

    RP = RecurrencePlot(x, metric="supremum", threshold=0., sparse_rqa=sparse)
    assert RP.max_diaglength() == 0
    assert RP.max_vertlength() == 0
    if not sparse:
        assert RP.max_white_vertlength() == RP.N

    RP = RecurrencePlot(x, metric="supremum", threshold=2., sparse_rqa=sparse)
    assert RP.max_diaglength() == (RP.N - 1)
    assert RP.max_vertlength() == RP.N
    if not sparse:
        assert RP.max_white_vertlength() == 0


def test_rqa_summary(small_RP):
    res = small_RP.rqa_summary()
    measures = ['RR', 'DET', 'L', 'LAM']
    assert all(res[m].dtype == DFIELD for m in measures)


def test_rqa_summary_numeric(small_RP_basic):
    res = small_RP_basic.rqa_summary()
    exp = {'RR': 0.48, 'DET': 0.8947, 'L': 8.4999, 'LAM': 0.9999}
    assert all(np.isclose(res[m], val, atol=1e-04)
               for m, val in exp.items())


@pytest.mark.parametrize('measure', ['diagline', 'vertline'])
@pytest.mark.parametrize('M', np.arange(5, 90, 40).tolist())
def test_resample_line_dist(measure: str, M: int, small_RP):
    res = getattr(small_RP, f"resample_{measure}_dist")(M)
    assert res.dtype == NODE
    assert res.shape[0] == small_RP.N
    assert (0 <= res).all()


@pytest.mark.parametrize(
    "var, exp", [("trapping", 4.7999), ("mean_recurrence", 3.9999)])
def test_time(small_RP_basic, var, exp):
    if small_RP_basic.sparse_rqa and var == "mean_recurrence":
        return
    res = getattr(small_RP_basic, f"{var}_time")()
    assert np.isclose(res, exp, atol=1e-04)


# test entropy

@pytest.mark.parametrize(
    "ts, measure, value",
    [(np.array([4, 7, 9, 10, 6, 11, 3]), m, v)
     for m, v in [("permutation", 0.5888), ("complexity", 0.29)]]
    + [(np.arange(20), "complexity", 0.0)])
def test_entropy(ts: np.ndarray, measure: str, value: float):
    rp = RecurrencePlot(ts[:, np.newaxis], threshold=1, dim=3, tau=1)
    assert np.isclose(getattr(rp, f"{measure}_entropy")(), value, atol=1e-04)


@pytest.mark.parametrize(
    'measure, exp',
    [('diag', 0.6931), ('vert', 1.2206), ('white_vert', 1.8848)])
def test_line_dist_entropy(measure: str, exp: float, small_RP_basic):
    if small_RP_basic.sparse_rqa and measure == "white_vert":
        return
    res = getattr(small_RP_basic, f"{measure}_entropy")()
    assert np.isclose(res, exp, atol=1e-04)



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
climate
=======

Provides classes for generating and analyzing complex climate networks.

Related Publications
~~~~~~~~~~~~~~~~~~~~
[Donges2009c]_, [Donges2009a]_, [Donges2009b]_, [Donges2011a]_, [Zou2011]_,
[Tominski2011]_, [Heitzig2012]_
"""

from ..core import GeoNetwork, GeoGrid, Network

from .climate_data import ClimateData
from .climate_network import ClimateNetwork
from .coupled_climate_network import CoupledClimateNetwork
from .coupled_tsonis import CoupledTsonisClimateNetwork
from .havlin import HavlinClimateNetwork
from .hilbert import HilbertClimateNetwork
from .map_plot import MapPlot
from .mutual_info import MutualInfoClimateNetwork
from .partial_correlation import PartialCorrelationClimateNetwork
from .rainfall import RainfallClimateNetwork
from .spearman import SpearmanClimateNetwork
from .tsonis import TsonisClimateNetwork
from .eventseries_climatenetwork import \
    EventSeriesClimateNetwork


#
#  Set global constants
#

#  Mean earth radius in kilometers
from ..core import EARTH_RADIUS



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for the analysis of dynamical systems and time series based
on recurrence plots, including measures of recurrence quantification
analysis (RQA) and recurrence network analysis.
"""

# array object and fast numerics
import numpy as np

from ..core import InteractingNetworks
from .recurrence_plot import RecurrencePlot
from .cross_recurrence_plot import CrossRecurrencePlot


#
#  Class definitions
#

class InterSystemRecurrenceNetwork(InteractingNetworks):

    """
    Generating and quantitatively analyzing inter-system recurrence networks.

    For a inter-system recurrence network, time series x and y do not need to
    have the same length! Formally, nodes are identified with state vectors in
    the common phase space of both time series. Hence, the time series need to
    have the same number of dimensions and identical physical units.
    Undirected links are added to describe recurrences within x and y as well
    as cross-recurrences between x and y. Self-loops are excluded in this
    undirected network representation.

    More information on the theory and applications of inter system recurrence
    networks can be found in [Feldhoff2012]_.

    **Examples:**

     - Create an instance of InterSystemRecurrenceNetwork with fixed
       recurrence thresholds and without embedding::

           InterSystemRecurrenceNetwork(x, y, threshold=(0.1, 0.2, 0.1))

     - Create an instance of InterSystemRecurrenceNetwork at a fixed
       recurrence rate and using time delay embedding::

           InterSystemRecurrenceNetwork(
               x, y, dim=3, tau=(2, 1), recurrence_rate=(0.05, 0.05, 0.02))
    """

    #
    #  Internal methods
    #

    def __init__(self, x, y, metric="supremum",
                 normalize=False, silence_level=0, **kwds):
        """
        Initialize an instance of InterSystemRecurrenceNetwork (ISRN).

        .. note::
           For an inter system recurrence network, time series x and y need to
           have the same number of dimensions!

        Creates an embedding of the given time series x and y, calculates a
        inter system recurrence matrix from the embedding and then creates
        an InteractingNetwork object from this matrix, interpreting the inter
        system recurrence matrix as the adjacency matrix of an undirected
        complex network.

        Either recurrence thresholds ``threshold`` or
        recurrence rates ``recurrence_rate`` have to be given as keyword
        arguments.

        Embedding is only supported for scalar time series. If embedding
        dimension ``dim`` and delay ``tau`` are **both** given as keyword
        arguments, embedding is applied. Multidimensional time series are
        processed as is by default.

        :type x: 2D Numpy array (time, dimension)
        :arg x: The time series x to be analyzed, can be scalar or
                multi-dimensional.
        :type y: 2D Numpy array (time, dimension)
        :arg y: The time series y to be analyzed, can be scalar or
                multi-dimensional.
        :type metric: tuple of string
        :arg metric: The metric for measuring distances in phase space
                     ("manhattan", "euclidean", "supremum").
        :arg bool normalize: Decide whether to normalize the time series to
                             zero mean and unit standard deviation.
        :arg int silence_level: The inverse level of verbosity of the object.
        :arg kwds: Additional options.
        :type threshold: tuple of number (three numbers)
        :keyword threshold: The recurrence threshold keyword for generating
                            the recurrence plot using fixed thresholds. Give
                            for each time series and the cross recurrence plot
                            separately.
        :type recurrence_rate: tuple of number (three numbers)
        :keyword recurrence_rate: The recurrence rate keyword for generating
                                  the recurrence plot using a fixed recurrence
                                  rate. Give separately for each time series.
        :keyword int dim: The embedding dimension. Must be the same for both
                          time series.
        :type tau: tuple of int
        :keyword tau: The embedding delay. Give separately for each time
                      series.
        """
        #  Store time series
        self.x = x.copy().astype("float32")
        """The time series x."""
        self.y = y.copy().astype("float32")
        """The time series y."""

        #  Reshape time series
        self.x.shape = (self.x.shape[0], -1)
        self.y.shape = (self.y.shape[0], -1)

        #  Get embedding dimension and delay from **kwds
        dim = kwds.get("dim")
        tau = kwds.get("tau")

        #  Check for consistency
        if self.x.shape[1] == self.y.shape[1]:
            #  Set silence_level
            self.silence_level = silence_level
            """The inverse level of verbosity of the object."""

            #  Get number of nodes in subnetwork x
            self.N_x = self.x.shape[0]
            """Number of nodes in subnetwork x."""

            #  Get number of nodes in subnetwork y
            self.N_y = self.y.shape[0]
            """Number of nodes in subnetwork y."""

            #  Get total number of nodes of ISRN
            self.N = self.N_x + self.N_y
            """Total number of nodes of ISRN."""

            #  Store type of metric
            self.metric = metric
            """The metric used for measuring distances in phase space."""

            #  Normalize time series
            if normalize:
                RecurrencePlot.normalize_time_series(self.x)
                RecurrencePlot.normalize_time_series(self.y)

            #  Embed time series if required
            self.dim = dim
            if dim is not None and tau is not None and self.x.shape[1] == 1:
                self.x_embedded = \
                    RecurrencePlot.embed_time_series(self.x, dim, tau[0])
                """The embedded time series x."""
                self.y_embedded = \
                    RecurrencePlot.embed_time_series(self.y, dim, tau[1])
                """The embedded time series y."""
            else:
                self.x_embedded = self.x
                self.y_embedded = self.y

            #  Get threshold or recurrence rate from **kwds, construct
            #  ISRN accordingly
            threshold = kwds.get("threshold")
            recurrence_rate = kwds.get("recurrence_rate")
            self.threshold = threshold

            if threshold is not None:
                #  Calculate the ISRN using the radius of neighborhood
                #  threshold
                ISRM = self.set_fixed_threshold(threshold)
            elif recurrence_rate is not None:
                #  Calculate the ISRN using a fixed recurrence rate
                ISRM = self.set_fixed_recurrence_rate(recurrence_rate)
            else:
                raise NameError("Please give either threshold or \
                                recurrence_rate to construct the joint \
                                recurrence plot!")

            InteractingNetworks.__init__(self, adjacency=ISRM, directed=False,
                                         silence_level=self.silence_level)
            #  No treatment of missing values yet!
            self.missing_values = False

        else:
            raise ValueError("Both time series x and y need to have the same \
                             dimension!")

    def __str__(self):
        """
        Returns a string representation.
        """
        return ("InterSystemRecurrenceNetwork: "
                f"time series shapes {self.x.shape}, {self.y.shape}.\n"
                f"Embedding dimension {self.dim if self.dim else 0}\n"
                f"Threshold {self.threshold}, {self.metric} metric.\n"
                f"{InteractingNetworks.__str__(self)}")

    #
    #  Service methods
    #

    def clear_cache(self):
        """
        Clean up memory by deleting information that can be recalculated from
        basic data.

        Extends the clean up methods of the parent classes.
        """
        #  Call clean up of RecurrencePlot objects
        self.rp_x.clear_cache()
        self.rp_y.clear_cache()

        #  Call clean up of CrossRecurrencePlot object
        self.crp_xy.clear_cache()

        #  Call clean up of InteractingNetworks
        InteractingNetworks.clear_cache(self)

    #
    #  Methods to handle inter system recurrence networks
    #

    def inter_system_recurrence_matrix(self):
        """
        Return the current inter system recurrence matrix :math:`ISRM`.

        :rtype: 2D square Numpy array
        :return: the current inter system recurrence matrix :math:`ISRM`.
        """
        #  Shortcuts
        N = self.N
        N_x = self.N_x
        N_y = self.N_y

        #  Init
        ISRM = np.zeros((N, N))

        #  Combine to inter system recurrence matrix
        ISRM[:N_x, :N_x] = self.rp_x.recurrence_matrix()
        ISRM[:N_x, N_x:N] = self.crp_xy.recurrence_matrix()
        ISRM[N_x:N, :N_x] = self.crp_xy.recurrence_matrix().transpose()
        ISRM[N_x:N, N_x:N] = self.rp_y.recurrence_matrix()

        return ISRM

    def set_fixed_threshold(self, threshold):
        """
        Create a inter system recurrence network at fixed thresholds.

        :type threshold: tuple of number (three numbers)
        :arg threshold: The three threshold parameters. Give for each
                        time series and the cross recurrence plot separately.
        """
        #  Compute recurrence matrices of x and y
        self.rp_x = RecurrencePlot(time_series=self.x_embedded,
                                   threshold=threshold[0],
                                   metric=self.metric,
                                   silence_level=self.silence_level)
        self.rp_y = RecurrencePlot(time_series=self.y_embedded,
                                   threshold=threshold[1],
                                   metric=self.metric,
                                   silence_level=self.silence_level)

        #  Compute cross-recurrence matrix of x and y
        self.crp_xy = CrossRecurrencePlot(x=self.x_embedded, y=self.y_embedded,
                                          threshold=threshold[2],
                                          metric=self.metric,
                                          silence_level=self.silence_level)

        #  Get combined ISRM
        ISRM = self.inter_system_recurrence_matrix()

        #  Set diagonal of ISRM to zero to avoid self-loops
        ISRM.flat[::self.N + 1] = 0
        return ISRM

    def set_fixed_recurrence_rate(self, density):
        """
        Create a inter system recurrence network at fixed link densities (
        recurrence rates).

        :type density: tuple of number (three numbers)
        :arg density: The three recurrence rate parameters. Give for each
                        time series and the cross recurrence plot separately.
        """
        #  Compute recurrence matrices of x and y
        self.rp_x = RecurrencePlot(time_series=self.x_embedded,
                                   recurrence_rate=density[0],
                                   metric=self.metric,
                                   silence_level=self.silence_level)
        self.rp_y = RecurrencePlot(time_series=self.y_embedded,
                                   recurrence_rate=density[1],
                                   metric=self.metric,
                                   silence_level=self.silence_level)

        #  Compute cross-recurrence matrix of x and y
        self.crp_xy = CrossRecurrencePlot(x=self.x_embedded, y=self.y_embedded,
                                          recurrence_rate=density[2],
                                          metric=self.metric,
                                          silence_level=self.silence_level)

        #  Get combined ISRM
        ISRM = self.inter_system_recurrence_matrix()

        #  Set diagonal of ISRM to zero to avoid self-loops
        ISRM.flat[::self.N + 1] = 0
        return ISRM

    #
    #  Methods to quantify inter system recurrence networks
    #

    def internal_recurrence_rates(self):
        """
        Return internal recurrence rates of subnetworks x and y.

        :rtype: tuple of number (float)
        :return: the internal recurrence rates of subnetworks x and y.
        """
        return (self.rp_x.recurrence_rate(),
                self.rp_y.recurrence_rate())

    def cross_recurrence_rate(self):
        """
        Return cross recurrence rate between subnetworks x and y.

        :rtype: number (float)
        :return: the cross recurrence rate between subnetworks x and y.
        """
        return self.crp_xy.cross_recurrence_rate()

    def cross_global_clustering_xy(self):
        """
        Return cross global clustering of x with respect to y.

        See [Feldhoff2012]_ for definition, further explanation and
        applications.

        :rtype: number (float)
        :return: the cross global clustering of x with respect to y.
        """
        return self.cross_global_clustering(np.arange(self.N_x),
                                            np.arange(self.N_x, self.N))

    def cross_global_clustering_yx(self):
        """
        Return cross global clustering of y with respect to x.

        See [Feldhoff2012]_ for definition, further explanation and
        applications.

        :rtype: number (float)
        :return: the cross global clustering of y with respect to x.
        """
        return self.cross_global_clustering(np.arange(self.N_x, self.N),
                                            np.arange(self.N_x))

    def cross_transitivity_xy(self):
        """
        Return cross transitivity of x with respect to y.

        See [Feldhoff2012]_ for definition, further explanation and
        applications.

        :rtype: number (float)
        :return: the cross transitivity of x with respect to y.
        """
        return self.cross_transitivity(np.arange(self.N_x),
                                       np.arange(self.N_x, self.N))

    def cross_transitivity_yx(self):
        """
        Return cross transitivity of y with respect to x.

        See [Feldhoff2012]_ for definition, further explanation and
        applications.

        :rtype: number (float)
        :return: the cross transitivity of y with respect to x.
        """
        return self.cross_transitivity(np.arange(self.N_x, self.N),
                                       np.arange(self.N_x))



# # Copyright (C) 2014 Paul Schultz and the SWIPO Project
#
# Authors (this file):
#   Stefan Schinkel <stefan.schinkel@gmail.com>
#   Paul Schultz pschultz@pik-potsdam.de>
#
#
# References: [Schultz2014]_, [Schultz2014a]_,
#
# ToDo:
#  - included Schultz' Predictor and growth algorithm (among others)


"""
Module contains class ResNetwork.

Provides function for computing resistance based networks. It is subclassed
from GeoNetwork and provides most GeoNetwork's functions/properties.

The class has the following instance variables::

    (bool) flagDebug     : flag for debugging mode
    (bool) flagComplex   : flag for complex input
    (ndarray) resistances: array of resistances (complex or real)

Overriden inherited methods::

    (str) __str__          : extended description
    (ndarray) get_adjacency: returns complex adjacency if needed
"""

#  Import NumPy for the array object and fast numerics
import numpy as np

# Import handler for sparse matrices
from scipy import sparse

#  Import iGraph for high performance graph theory tools written in pure ANSI-C
import igraph

from ._ext.types import to_cy, FIELD, DFIELD
from ._ext.numerics import _vertex_current_flow_betweenness, \
    _edge_current_flow_betweenness

# Import things we inherit from
from .geo_network import GeoNetwork
from .geo_grid import GeoGrid

# a network Error (use uncertain)
# from .network import NetworkError


class ResNetwork(GeoNetwork):
    """ A resistive network class

    ResNetwork, provides methods for an extended analysis of
    resistive/resistance-based networks.

    **Examples:**

    >>> print(ResNetwork.SmallTestNetwork())
    ResNetwork:
    GeoNetwork:
    Network: undirected, 5 nodes, 5 links, link density 0.500.
    Geographical boundaries:
             time     lat     lon
       min    0.0    0.00 -180.00
       max    9.0   90.00  180.00
    Average resistance: 2.4
    """
###############################################################################
# ##                       MAGIC FUNCTIONS                                 ## #
###############################################################################

    def __init__(self, resistances, grid=None, adjacency=None, edge_list=None,
                 directed=False, node_weight_type=None, silence_level=2):

        """Initialize an instance of ResNetwork.

        :type resistances: 2D NumPy array
        :arg resistances: A matrix with the resistances

        :type grid: GeoGrid object
        :arg grid: The GeoGrid object describing the network's spatial
            embedding.

        :type adjacency: 2D NumPy array (int8) [index, index]
        :arg adjacency: The network's adjacency matrix.

        :type edge_list: array-like list of lists
        :arg  edge_list: Edge list of the new network.
                         Entries [i,0], [i,1] contain the end-nodes of an edge.

        :type directed: boolean
        :arg directed: Determines, whether the network is treated as directed.

        :type node_weight_type: string
        :arg node_weight_type: The type of geographical node weight to be used.

        :type silence_level: number (int)
        :arg silence_level: The inverse level of verbosity of the object.
        """

        # 1) prepare params so we can init() the parent

        # 1a) an adjacency matrix
        #     by default, if only resistance are given,
        #     these define the adjacency
        if adjacency is None:
            if silence_level < 2:
                print("Using adjacency as definded by resistances")

            adjacency = np.zeros(resistances.shape)
            adjacency[resistances != 0] = 1

        # 1b) a Grid object
        #     an actual grid might not exist, so we fake one
        if grid is None:
            if silence_level < 2:
                print("Using dummy grid")
            grid = GeoGrid(
                time_seq=np.arange(10), lat_seq=np.absolute(
                    np.linspace(-90, 90, adjacency.shape[0])),
                lon_seq=np.linspace(-180, 180, adjacency.shape[0]),
                silence_level=0)

        # 2) init parent
        GeoNetwork.__init__(self, grid, adjacency=adjacency,
                            node_weight_type=node_weight_type,
                            silence_level=silence_level)

        # 3) add ResNetwork specific parts

        # 3a) set the resitance values
        #     this sets the property and forces the
        #     updating of the admittance and R
        self.sparse_Adm = None
        self.adm_graph = None
        self.sparse_R = None
        self.update_resistances(resistances)

        # 4) cache
        self._effective_resistances = None

    def __str__(self):
        """
        Return a short summary of the resistive network.
        """
        return (f"ResNetwork:\n{GeoNetwork.__str__(self)}\n"
                f"Average resistance: {self.resistances.mean()}")

###############################################################################
# ##                       PUBLIC FUNCTIONS                                ## #
###############################################################################
    @staticmethod
    def SmallTestNetwork():
        r"""
        Create a small test network with unit resistances of the following
        topology::

            0------1--------3------4
                    \      /
                     \    /
                      \  /
                       \/
                       2

        :rtype: Resistive Network instance
        :return: an ResNetwork instance for testing purposes.

        **Examples:**

        >>> res = ResNetwork.SmallTestNetwork()
        >>> isinstance(res, ResNetwork)
        True
        """
        adjacency = np.array([[0, 1, 0, 0, 0],
                              [1, 0, 1, 1, 0],
                              [0, 1, 0, 1, 0],
                              [0, 1, 1, 0, 1],
                              [0, 0, 0, 1, 0]], dtype='int8')

        # sample symmetric resistances w/ rational reciprocals
        # resistances = np.array([2,2,8,2,8,8,2,8,10,10])

        # # the resistances should be a full matrix
        resistances = np.array([[0, 2, 0, 0, 0],
                                [2, 0, 8, 2, 0],
                                [0, 8, 0, 8, 0],
                                [0, 2, 8, 0, 10],
                                [0, 0, 0, 10, 0]])
        # a grid
        grid = GeoGrid(
            time_seq=np.arange(10), lat_seq=np.absolute(
                np.linspace(-90, 90, adjacency.shape[0])),
            lon_seq=np.linspace(-180, 180, adjacency.shape[0]),
            silence_level=0)

        return ResNetwork(resistances, grid=grid, adjacency=adjacency)

    @staticmethod
    def SmallComplexNetwork():
        """
        A test network with complex resistances analogue to
        SmallTestNetwork()

        :rtype: Resistive Network instance
        :return: an ResNetwork instance with complex resistances

        **Examples:**

        >>> res = ResNetwork.SmallComplexNetwork()
        >>> isinstance(res, ResNetwork)
        True
        >>> res.flagComplex
        True
        >>> adm = res.get_admittance()
        >>> print(adm.real)
        [[ 0.      0.1     0.      0.      0.    ]
         [ 0.1     0.      0.0625  0.25    0.    ]
         [ 0.      0.0625  0.      0.0625  0.    ]
         [ 0.      0.25    0.0625  0.      0.05  ]
         [ 0.      0.      0.      0.05    0.    ]]

        >>> print(adm.imag)
        [[ 0.     -0.2     0.      0.      0.    ]
         [-0.2     0.     -0.0625 -0.25    0.    ]
         [ 0.     -0.0625  0.     -0.0625  0.    ]
         [ 0.     -0.25   -0.0625  0.     -0.05  ]
         [ 0.      0.      0.     -0.05    0.    ]]

        """

        resistances = np.zeros((5, 5), dtype=complex)
        resistances.real = [[0, 2, 0, 0, 0],
                            [2, 0, 8, 2, 0],
                            [0, 8, 0, 8, 0],
                            [0, 2, 8, 0, 10],
                            [0, 0, 0, 10, 0]]
        resistances.imag = [[0, 4, 0, 0, 0],
                            [4, 0, 8, 2, 0],
                            [0, 8, 0, 8, 0],
                            [0, 2, 8, 0, 10],
                            [0, 0, 0, 10, 0]]

        return ResNetwork(resistances)

    def update_resistances(self, resistances):
        """
        Update the resistance matrix

        This function is called to changed the resistance matrix. It sets the
        property and the calls the :meth:`update_admittance` and
        :meth:`update_R` functions.

        :rtype: None

        **Examples:**

        >>> # test network with given resistances
        >>> res = ResNetwork.SmallTestNetwork()
        >>> print(res.resistances)
        [[ 0  2  0  0  0]
         [ 2  0  8  2  0]
         [ 0  8  0  8  0]
         [ 0  2  8  0 10]
         [ 0  0  0 10  0]]
        >>> # print admittance and admittance Laplacian
        >>> print(res.get_admittance())
        [[ 0.     0.5    0.     0.     0.   ]
         [ 0.5    0.     0.125  0.5    0.   ]
         [ 0.     0.125  0.     0.125  0.   ]
         [ 0.     0.5    0.125  0.     0.1  ]
         [ 0.     0.     0.     0.1    0.   ]]
        >>> print(res.admittance_lapacian())
        [[ 0.5   -0.5    0.     0.     0.   ]
         [-0.5    1.125 -0.125 -0.5    0.   ]
         [ 0.    -0.125  0.25  -0.125  0.   ]
         [ 0.    -0.5   -0.125  0.725 -0.1  ]
         [ 0.     0.     0.    -0.1    0.1  ]]
        >>> # now update to unit resistance
        >>> res.update_resistances(res.adjacency)
        >>> # and check new admittance/admittance Laplacian
        >>> print(res.get_admittance())
        [[ 0.  1.  0.  0.  0.]
         [ 1.  0.  1.  1.  0.]
         [ 0.  1.  0.  1.  0.]
         [ 0.  1.  1.  0.  1.]
         [ 0.  0.  0.  1.  0.]]
        >>> print(res.admittance_lapacian())
        [[ 1. -1.  0.  0.  0.]
         [-1.  3. -1. -1.  0.]
         [ 0. -1.  2. -1.  0.]
         [ 0. -1. -1.  3. -1.]
         [ 0.  0.  0. -1.  1.]]
        """

        # ensure ndarray
        if not isinstance(resistances, np.ndarray):
            resistances = np.array(resistances)

        # check complex/real
        self.flagComplex = np.iscomplexobj(resistances)

        # set property
        self.resistances = resistances

        # update the admittance
        self.update_admittance()

        # and update R
        self.update_R()

    def update_admittance(self):
        """
        Updates admittance matrix which is inverse the resistances

        :rtype: none

        **Examples:**

        >>> res = ResNetwork.SmallTestNetwork();print(res.get_admittance())
        [[ 0.     0.5    0.     0.     0.   ]
         [ 0.5    0.     0.125  0.5    0.   ]
         [ 0.     0.125  0.     0.125  0.   ]
         [ 0.     0.5    0.125  0.     0.1  ]
         [ 0.     0.     0.     0.1    0.   ]]
        >>> print(type(res.get_admittance()))
        <class 'numpy.ndarray'>
        """
        # a sparse matrix for the admittance values
        # we start w/ a lil_matrix, maybe convert that
        # to csr_matrix ( Compressed Sparse Row) or
        # to csc_matrix ( Compressed Sparse Column) later

        # complex number support
        if self.flagComplex:
            dtype = complex
        else:
            dtype = float

        self.sparse_Adm = sparse.lil_matrix((self.N, self.N), dtype=dtype)

        # get the edges
        edgeList = list(self.edge_list())

        # populate array
        for edge in edgeList:
            # print("setting %d %d to %f" % (
            #     edge[0], edge[1], 1./self.resistances[edge[0], edge[1]]))
            self.sparse_Adm[edge[0], edge[1]] = \
                1./self.resistances[edge[0], edge[1]]

        # Similar to GeoNetwork, we embed an iGraph instance for
        # the admittance matrix
        self.adm_graph = igraph.Graph(n=self.N, edges=edgeList,
                                      directed=self.directed)
        self.graph.simplify()

    def get_admittance(self):
        """Return the (possibly non-symmetric) dense admittance matrix

        :rtype: square NumPy matrix [node,node] of ints

        **Examples:**

        >>> res = ResNetwork.SmallTestNetwork();print(res.get_admittance())
        [[ 0.     0.5    0.     0.     0.   ]
         [ 0.5    0.     0.125  0.5    0.   ]
         [ 0.     0.125  0.     0.125  0.   ]
         [ 0.     0.5    0.125  0.     0.1  ]
         [ 0.     0.     0.     0.1    0.   ]]
        >>> print(type( res.get_admittance() ))
        <class 'numpy.ndarray'>
        """
        return self.sparse_Adm.toarray()

    def update_R(self):
        """Updates R, the pseudo inverse of the admittance Laplacian

        This function is run, whenever the admittance is changed.

        :rtype: none

        **Examples:**

        >>> res = ResNetwork.SmallTestNetwork();print(res.get_admittance())
        [[ 0.     0.5    0.     0.     0.   ]
         [ 0.5    0.     0.125  0.5    0.   ]
         [ 0.     0.125  0.     0.125  0.   ]
         [ 0.     0.5    0.125  0.     0.1  ]
         [ 0.     0.     0.     0.1    0.   ]]
        >>> print(type( res.get_admittance() ))
        <class 'numpy.ndarray'>

        """
        # a sparse matrix for the admittance values
        self.sparse_R = sparse.lil_matrix(
            np.linalg.pinv(self.admittance_lapacian()))

    def get_R(self):
        """Return the pseudo inverse of of the admittance Laplacian

        The pseudoinverse is used of the novel betweeness  measures such as
        :meth:`vertex_current_flow_betweenness` and
        :meth:`edge_current_flow_betweenness` It is computed on instantiation
        and on change of the resistances/admittance

        :returns: the pseudoinverse of the admittange Laplacian
        :rtype: ndarray (float)

        **Examples:**

        >>> res = ResNetwork.SmallTestNetwork();print(res.get_R())
        [[ 2.28444444  0.68444444 -0.56       -0.20444444 -2.20444444]
         [ 0.68444444  1.08444444 -0.16        0.19555556 -1.80444444]
         [-0.56       -0.16        3.04       -0.16       -2.16      ]
         [-0.20444444  0.19555556 -0.16        1.08444444 -0.91555556]
         [-2.20444444 -1.80444444 -2.16       -0.91555556  7.08444444]]
        """
        return self.sparse_R.toarray()

    def admittance_lapacian(self):
        """
        Return the (possibly non-symmetric) dense Laplacian matrix of the
        admittance.

        :rtype: square NumPy matrix [node,node] of

        **Examples:**

        >>> print(ResNetwork.SmallTestNetwork().admittance_lapacian())
        [[ 0.5   -0.5    0.     0.     0.   ]
         [-0.5    1.125 -0.125 -0.5    0.   ]
         [ 0.    -0.125  0.25  -0.125  0.   ]
         [ 0.    -0.5   -0.125  0.725 -0.1  ]
         [ 0.     0.     0.    -0.1    0.1  ]]
        >>> print(type( ResNetwork.SmallTestNetwork().admittance_lapacian() ))
        <class 'numpy.ndarray'>
         """

        return np.diag(sum(self.get_admittance())) - self.get_admittance()

    def admittive_degree(self):
        """admittive degree of the network

        The admittive (or effective) degree of the resistive network,
        which is the counterpart to the traditional degree.

        :rtype: 1D NumPy array

        **Examples:**

        >>> print(ResNetwork.SmallTestNetwork().admittive_degree())
        [ 0.5    1.125  0.25   0.725  0.1  ]
        >>> print(type( ResNetwork.SmallTestNetwork().admittive_degree() ))
        <class 'numpy.ndarray'>
        """
        return np.sum(self.get_admittance(), axis=0)

    def average_neighbors_admittive_degree(self):
        """ Average neighbour effective degree

        :rtype: 1D NumPy array

        **Examples:**

        >>> print(ResNetwork.SmallTestNetwork().\
                average_neighbors_admittive_degree())
        [ 2.25  1.31111111  7.4  2.03448276  7.25  ]
        >>> print(type(ResNetwork.SmallTestNetwork().admittive_degree()))
        <class 'numpy.ndarray'>
        """

        # get the admittive degree (row sum)
        # and adjacency matrix
        ad = self.admittive_degree()
        adj = self.adjacency

        # in case of complex resistances we also use the dot product,
        # but of complex-valued arrays
        # This is NOT right, BUT:
        # np.dot treats complex numbers wrongly and computes the
        # dot product of the real and the imag part seperately
        # which in our case is exactly what we want
        if self.flagComplex:
            adj = np.array(adj, dtype=complex)
            adj.imag = adj.real

        # dot product of adjacency and degree
        # normalised by the row sum (admittive degree)
        return np.dot(adj, ad) / ad

        # to sweave later on:
        # N = self.N
        # adjacency = self.adjacency
        # ANED = np.zeros(N)
        # ED = self.admittive_degree()
        # for i in range(N):
        #     sum = 0
        #     for j in range(N):
        #         sum += adjacency[i][j]*ED[j]
        #     ANED[i] = sum/ED[i]
        # # print(ANED)

    def local_admittive_clustering(self):
        r"""
        Return node wise admittive clustering coefficient (AC).

        The AC is the electrical analogue of the clustering coefficient for
        regular network (see :meth:`.get_admittive_ws_clustering` and
        :meth:`.get_local_clustering` and sometimes called Effective Clustering
        (EC))

        The admittive clustering (:math:`ac`) of node :math:`i` is defined as:

        .. math::
            \text{ac}_i = \frac
                {\sum_{j,k}^N\alpha_{i,j},\alpha_{i,k},\alpha_{j,k}}
                {\text{ad}_i(\text{d}_i-1)}

        where
            - :math:`\alpha` is the admittance matrix
            - :math:`ad_i` is the admittive degree of the node :math:`i`
            - :math:`d_i` is the degree of the node :math:`i`

        :rtype: 1d NumPy array (float)

        **Examples:**

        >>> res =  ResNetwork.SmallTestNetwork()
        >>> print(res.local_admittive_clustering())
        [ 0.  0.00694444  0.0625  0.01077586  0. ]
        >>> print(type(res.local_admittive_clustering()))
        <class 'numpy.ndarray'>
        """

        # needed vals: admittance matrix and admittiv_degree
        # are already complex/real as needed
        admittance = self.get_admittance()
        ad = self.admittive_degree()

        # output and the degree have to be switched
        if self.flagComplex:
            d = np.array(self.degree(), dtype=complex)
            ac = np.zeros(self.N, dtype=complex)

        else:
            d = self.degree()
            ac = np.zeros(self.N)

        # TODO: Sweave me!
        for i in range(self.N):
            dummy = 0
            for j in range(self.N):
                for k in range(self.N):
                    dummy += admittance[i][j]*admittance[i][k]*admittance[j][k]
            if d[i] == 1:
                ac[i] = 0
            else:
                ac[i] = dummy/(ad[i] * (d[i]-1))

        # return
        return ac

    def global_admittive_clustering(self):
        """
        Return node wise admittive clustering coefficient.

        :rtype: NumPy float

        **Examples:**

        >>> res =  ResNetwork.SmallTestNetwork()
        >>> print("%.3f" % res.global_admittive_clustering())
        0.016
        >>> print(type(res.global_admittive_clustering()))
        <class 'numpy.float64'>
        """

        return self.local_admittive_clustering().mean()

    def effective_resistance(self, a, b):
        """
        Return the effective resistance (ER) between two nodes
        a and b. The ER is the electrical analogue to the shortest path
        where a is considered as "source" and b as the "sink"

        :type a: int
        :arg a: index of the "source" node

        :type b: int
        :arg b: index of the "sink" node

        :rtype: NumPy float

        **Examples:**

        >>> res = ResNetwork.SmallTestNetwork()
        >>> print(res.effective_resistance(1,1))
        0.0
        >>> print(type( res.effective_resistance(1,1) ))
        <class 'float'>
        >>> print("%.3f" % res.effective_resistance(1,2))
        4.444
        >>> print(type( res.effective_resistance(1,1) ))
        <class 'float'>
        """
        # return def for self-loop
        if a == b:
            if self.flagComplex:
                return complex(0.0)
            else:
                return float(0.0)

        # Get pseudoinverse of the Laplacian
        R = self.get_R()

        # return looked-up values
        return R[a, a] - R[a, b] - R[b, a] + R[b, b]

    def average_effective_resistance(self):
        """
        Return the average effective resistance (<ER>) of the resistive
        network, the average resistances for all "paths" (connections)

        :rtype: float

        **Examples:**

        >>> res = ResNetwork.SmallTestNetwork()
        >>> print("%.5f" % res.average_effective_resistance())
        7.28889
        >>> print(type( res.average_effective_resistance() ))
        <class 'numpy.float64'>
        """

        # since the NW is symmetric, we can only
        # sum over upper triangle, excluding zeros
        # but multiply by 2 later on

        # we also store a hidden, quick access var
        self._effective_resistances = np.array([])

        for i in range(self.N):
            for j in range(i):
                self._effective_resistances = np.append(
                    self._effective_resistances,
                    self.effective_resistance(i, j))

        return 2*np.sum(self._effective_resistances) / (self.N*(self.N-1))

    def diameter_effective_resistance(self):
        """
        Return the diameter (the highest resistance path between any nodes).

        :rtype: float

        **Examples:**

        >>> res = ResNetwork.SmallTestNetwork()
        >>> print("%.3f" % res.diameter_effective_resistance())
        Re-computing all effective resistances
        14.444
        >>> print(type(res.diameter_effective_resistance()))
        <class 'numpy.float64'>
        >>> res = ResNetwork.SmallTestNetwork()
        >>> x = res.average_effective_resistance()
        >>> print("%.3f" % res.diameter_effective_resistance())
        14.444
        """
        # try to use pre-computed values
        if self._effective_resistances is not None:
            diameter = np.max(self._effective_resistances)
        else:
            print("Re-computing all effective resistances")
            self.average_effective_resistance()
            diameter = np.max(self._effective_resistances)

        return diameter

    def effective_resistance_closeness_centrality(self, a):
        """
        The effective resistance closeness centrality (ERCC) of node a

        :type a: int
        :arg a: index of the "source" node

        :rtype: NumPy float

        **Examples:**

        >>> res = ResNetwork.SmallTestNetwork()
        >>> print("%.3f" % res.effective_resistance_closeness_centrality(0))
        0.154
        >>> print("%.3f" % res.effective_resistance_closeness_centrality(4))
        0.080
        """

        # alloc
        ERCC = DFIELD(0.0)

        # compute
        for i in range(self.N):
            ERCC += self.effective_resistance(a, i)
        # ERCC /=  np.square( self.N - 1 )
        ERCC = (self.N - 1) / ERCC

        # return
        return ERCC

    def vertex_current_flow_betweenness(self, i):
        r"""
        Vertex Current Flow Betweeness (VCFB) of a node i.

        The electrial version of Newmann's node betweeness is here
        defined as the Vertex Current Flow Betweeness (VCGB) of a node

        .. math::
            VCFB_i := \frac{ 2 }{ n \left( n-1 \right)} \sum_{s<t} I_i^{st}

        where

        .. math::
            I_i^{st} &= \frac{1}{2}\sum_{j} \Gamma_{i,j} | V_i - V_j |\\
                     &= \frac{1}{2}\sum_{j} \Gamma_{i,j}
                     | I_s(R_{i,s}-R_{j,s}) + I_t(R_{j,t}-R_{i,t}) |

        and further:
            - :math:`I_{s}^{st} := I_{s}`
            - :math:`I_{t}^{st}  := I_{t}`
            - :math:`\Gamma` is the admittance matrix
            - :math:`R` is the pseudoinverse of the admittance Laplacian

        :arg int a: index of the "source" node
        :rtype: NumPy float

        **Examples:**

        >>> res = ResNetwork.SmallTestNetwork()
        >>> print("%.3f" % res.vertex_current_flow_betweenness(1))
        0.389
        >>> print("%.3f" % res.vertex_current_flow_betweenness(2))
        0.044
        """
        # set params
        Is = It = FIELD(1.0)
        return _vertex_current_flow_betweenness(
            self.N, Is, It,
            to_cy(self.get_admittance(), FIELD), to_cy(self.get_R(), FIELD), i)

    def edge_current_flow_betweenness(self):
        """The electrial version of Newmann's edge betweeness

        :rtype: NumPy float

        **Examples:**

        >>> res = ResNetwork.SmallTestNetwork()
        >>> print(r(res.edge_current_flow_betweenness()))
        [[ 0.      0.4     0.      0.      0.    ]
         [ 0.4     0.      0.2444  0.5333  0.    ]
         [ 0.      0.2444  0.      0.2444  0.    ]
         [ 0.      0.5333  0.2444  0.      0.4   ]
         [ 0.      0.      0.      0.4     0.    ]]
        >>> # update to unit resistances
        >>> res.update_resistances(res.adjacency)
        >>> print(r(res.edge_current_flow_betweenness()))
        [[ 0.      0.4     0.      0.      0.    ]
         [ 0.4     0.      0.3333  0.4     0.    ]
         [ 0.      0.3333  0.      0.3333  0.    ]
         [ 0.      0.4     0.3333  0.      0.4   ]
         [ 0.      0.      0.      0.4     0.    ]]
        """
        # set currents
        Is = It = FIELD(1)

        return _edge_current_flow_betweenness(
            self.N, Is, It,
            to_cy(self.get_admittance(), FIELD), to_cy(self.get_R(), FIELD))


###############################################################################
# ##                       FUNCTIONS ATTIC                                 ## #
###############################################################################

    # These functions are no longer needed as the computation can be broken
    # down to some indexing but they are kept as implementatin references
    # since all git logs will be lost when adding resnw to pyunicorn.
    #
    #
    # def _effective_resistance_python(self,a,b):
    #     """Python version of the effective resistance
    #     """
    #     # Get pseudoinverse of the Laplacian
    #     R = self.get_R()

    #     t = R[a,a] - R[a,b] - R[b,a] + R[b,b]
    #     print("the t is %f + %fi " % (t.real,t.imag))
    #     # construct a vector that is all zero except for
    #     # the source (a) and the sink (b)
    #     # if self.flagComplex:
    #     #     base = np.zeros(self.N,dtype=complex)
    #     # else:
    #     base = np.zeros(self.N)

    #     base[a] = 1;
    #     base[b] = -1;

    #     # multiply every row of R with the above vector
    #     # and sum across rows => np.sum(R * base, axis=1) and then
    #     # build the scalar product of the result and the base vector
    #     ER = np.dot( np.sum(R * base, axis=1), base )

    #     return ER

    # def _effective_resistance_weave(self,a,b):
    #     """ C version of effective resistance
    #     """
    #     # Get pseudoinverse of the Laplacian
    #     R = self.get_R()
    #     N = np.float( self.N)

    #     code = \
    #     """
    #     int i=0;
    #     int j=0;
    #     double ER = 0.0;

    #     // vector for temp values
    #     double tmp[int(N)];

    #     // setup the base vector
    #     double base[int(N)];

    #     base[a] = 1;
    #     base[b] = -1;

    #     for(i=0;i<N; i++){
    #         for (j=0; j<N; j++){
    #             tmp[i] += R2(i,j) * base[j];
    #         } //for j
    #         ER += base[i] * tmp[i];
    #     } // for i

    #     return_val = ER;
    #     """
    #     variables = ['a','b','N','R'];
    #     ER = weave.inline(code,variables,compiler = "gcc")
    #     return ER

    # def _effective_resistance_weave_complex(self,a,b):
    #     """ C version of effective resistance
    #     """
    #     # Get pseudoinverse of the Laplacian
    #     R = self.get_R()
    #     N = np.float( self.N)

    #     code = \
    #     """
    #     int i=0;
    #     int j=0;
    #     std::complex<double> ER = 0.0;

    #     // vector for base and values
    #     std::complex<double> base[int(N)];
    #     base[a].real() = 1;
    #     base[b].real() = -1;

    #     // vector for temp values
    #     std::complex<double> tmp[int(N)];

    #     /*
    #     tmp = (double*) calloc(N,sizeof(double));

    #    /// handling of complex numbers via struct definition
    #     // struct describing complex no.
    #      typedef struct {
    #         double real;
    #         double imag;
    #     } complex_def;

    #     complex_def *base;

    #     base = calloc(N * sizeof(*array));

    #     */

    #     for(i=0;i<N; i++){
    #         for (j=0; j<N; j++){
    #             tmp[i] += R2(i,j) * base[j];
    #         } //for j
    #         ER += base[i] * tmp[i];
    #     } // for i

    #     return_val = ER;
    #     """
    #     variables = ['a','b','N','R'];
    #     ER = weave.inline(code,variables,compiler = "gcc",
    #         headers=["<complex.h>"])
    #     return ER



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for generating and analyzing complex climate networks.
"""

from typing import Tuple
from collections.abc import Hashable

import numpy as np
try:
    import scipy.signal
except ImportError:
    print("climate: Package scipy.signal could not be loaded. Some "
          "functionality in class HilbertClimateNetwork might not be "
          "available!")

from .climate_data import ClimateData
from .climate_network import ClimateNetwork


#
#  Define class HilbertClimateNetwork
#

class HilbertClimateNetwork(ClimateNetwork):

    """
    Encapsulates a Hilbert climate network.

    The associated similarity matrix is based on Hilbert coherence between
    time series.

    Hilbert climate networks can be directed and undirected. Optional
    directionality is based on the average phase difference between time
    series.

    A preliminary study of Hilbert climate networks is presented in
    [Donges2009c]_.
    """

    #
    #  Defines internal methods
    #

    def __init__(self, data, threshold=None, link_density=None,
                 non_local=False, directed=True, node_weight_type="surface",
                 silence_level=0):
        """
        Initialize an instance of HilbertClimateNetwork.

        .. note::
           Either threshold **OR** link_density have to be given!

        Possible choices for ``node_weight_type``:
          - None (constant unit weights)
          - "surface" (cos lat)
          - "irrigation" (cos**2 lat)

        :type data: :class:`.ClimateData`
        :arg data: The climate data used for network construction.
        :arg float threshold: The threshold of similarity measure, above which
            two nodes are linked in the network.
        :arg float link_density: The networks's desired link density.
        :arg bool non_local: Determines, whether links between spatially close
            nodes should be suppressed.
        :arg bool directed: Determines, whether the network is constructed as
            directed.
        :arg str node_weight_type: The type of geographical node weight to be
            used.
        :arg int silence_level: The inverse level of verbosity of the object.
        """
        if silence_level <= 1:
            print("Generating a Hilbert climate network...")
        self.silence_level = silence_level

        #  Set instance variables
        self._coherence_phase = None
        assert isinstance(data, ClimateData)
        self.data: ClimateData = data
        """The climate data used for network construction."""
        self.N = data.grid.N
        self._threshold = threshold
        self._prescribed_link_density = link_density

        self._set_directed(directed, calculate_coherence=True)
        ClimateNetwork.__init__(self, grid=self.data.grid,
                                similarity_measure=self._similarity_measure,
                                threshold=threshold,
                                link_density=link_density,
                                non_local=non_local,
                                directed=directed,
                                node_weight_type=node_weight_type,
                                silence_level=silence_level)
        self._set_directed(directed, calculate_coherence=False)

    def __cache_state__(self) -> Tuple[Hashable, ...]:
        return ClimateNetwork.__cache_state__(self)

    def __rec_cache_state__(self) -> Tuple[object, ...]:
        return ClimateNetwork.__rec_cache_state__(self) + (self.data,)

    def __str__(self):
        """
        Return a string representation.
        """
        return 'HilbertClimateNetwork:\n' + ClimateNetwork.__str__(self)

    def clear_cache(self):
        """
        Clean up cache.
        """
        try:
            del self._coherence_phase
        except AttributeError:
            pass

    #
    #  Defines methods to calculate Hilbert correlation measures
    #

    def _set_directed(self, directed, calculate_coherence=True):
        """
        Switch between directed and undirected Hilbert climate network.

        :arg bool directed: Determines whether the network is constructed as
            directed.
        :arg bool calculate_coherence: Determines whether coherence and phase
            are calculated from data or the directed adjacency matrix is
            constructed from coherence and phase information.
        """
        if calculate_coherence:
            results = self._calculate_hilbert_correlation(self.data.anomaly())
            self._similarity_measure = results[0]
            self._coherence_phase = results[1]
            self.directed = directed
        else:
            # The phase is only used for directed Hilbert networks.
            if directed:
                self.adjacency = self.adjacency * (self.phase_shift() > 0)

    def set_directed(self, directed):
        """
        Switch between directed and undirected Hilbert climate network.

        Also performs the complete network generation.

        :arg bool directed: Determines whether the network is constructed as
            directed.
        """
        self._set_directed(directed, calculate_coherence=True)
        self._regenerate_network()
        self._set_directed(directed, calculate_coherence=False)

    def _calculate_hilbert_correlation(self, anomaly):
        """
        Calculate Hilbert coherence and phase matrices.

        Output corresponds to modulus and argument of the complex correlation
        coefficients between all pairs of analytic signals calculated from
        anomaly data, as described in [Bergner2008]_.

        :type anomaly: 2D Numpy array [time, index]
        :arg anomaly: The anomaly data for network construction.

        :rtype: tuple of two 2D Numpy matrices [index, index]
        :return: the Hilbert coherence and phase matrices.
        """
        if self.silence_level <= 1:
            print("Calculating Hilbert transform correlation measures "
                  "following [Bergner2008]_...")

        #  Calculate the analytic signals associated with the anomaly time
        #  series.
        analytic_signals = np.apply_along_axis(scipy.signal.hilbert, 0,
                                               anomaly)

        #  Normalize analytic signal time series to zero mean and unit variance
        self.data.normalize_time_series_array(analytic_signals)

        #  Get length of time series
        n_time = anomaly.shape[0]

        #  Calculate matrix of complex correlation coefficients
        complex_corr = np.dot(analytic_signals.transpose(),
                              analytic_signals.conjugate()) / float(n_time - 1)

        #  Clean up
        del analytic_signals

        #  Calculate the coherence, i.e. the modulus of the complex
        #  correlation coefficient
        coherence = np.abs(complex_corr)

        #  Calculate the average phase between signals
        phase = np.angle(complex_corr)

        return (coherence, phase)

    def coherence(self):
        """
        Return the Hilbert coherence matrix.

        :rtype: 2D Numpy array [index, index]
        :return: the Hilbert coherence matrix.
        """
        return self.similarity_measure()

    def phase_shift(self):
        """
        Return the average phase shift matrix.

        :rtype: 2D Numpy array [index, index]
        :return: the average phase shift matrix.
        """
        return self._coherence_phase



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for the analysis of dynamical systems and time series based
on recurrence plots, including measures of recurrence quantification
analysis (RQA) and recurrence network analysis.
"""

# array object and fast numerics
import numpy as np

from ..core import Network
from .joint_recurrence_plot import JointRecurrencePlot


#
#  Class definitions
#

class JointRecurrenceNetwork(JointRecurrencePlot, Network):

    """
    Class JointRecurrenceNetwork for generating and quantitatively analyzing
    joint recurrence networks.

    For a joint recurrence network, time series x and y need to have the same
    length! Formally, nodes are identified with sampling points in time, while
    an undirected link (i,j) is introduced if x at time i is recurrent to x at
    time j and also y at time i is recurrent to y at time j. Self-loops are
    excluded in this undirected network representation.

    More information on the theory and applications of joint recurrence
    networks can be found in [Feldhoff2013]_.

    **Examples:**

     - Create an instance of JointRecurrenceNetwork with a fixed recurrence
       threshold and without embedding::

           JointRecurrenceNetwork(x, y, threshold=(0.1,0.2))

     - Create an instance of JointRecurrenceNetwork with a fixed recurrence
       threshold in units of STD and without embedding::

           JointRecurrenceNetwork(x, y, threshold_std=(0.03,0.05))

     - Create an instance of JointRecurrenceNetwork at a fixed recurrence rate
       and using time delay embedding::

           JointRecurrenceNetwork(
               x, y, dim=(3,5), tau=(2,1),
               recurrence_rate=(0.05,0.04)).recurrence_rate()
    """

    #
    #  Internal methods
    #

    def __init__(self, x, y, metric=("supremum", "supremum"),
                 normalize=False, lag=0, silence_level=0, **kwds):
        """
        Initialize an instance of JointRecurrenceNetwork.

        .. note::
           For a joint recurrence network, time series x and y need to have the
           same length!

        Creates an embedding of the given time series x and y, calculates a
        joint recurrence plot from the embedding and then creates a Network
        object from the joint recurrence plot, interpreting the joint
        recurrence matrix as the adjacency matrix of an undirected complex
        network.

        Either recurrence thresholds ``threshold``/``threshold_std`` or
        recurrence rates ``recurrence_rate`` have to be given as keyword
        arguments.

        Embedding is only supported for scalar time series. If embedding
        dimension ``dim`` and delay ``tau`` are **both** given as keyword
        arguments, embedding is applied. Multidimensional time series are
        processed as is by default.

        :type x: 2D Numpy array (time, dimension)
        :arg x: The time series x to be analyzed, can be scalar or
            multi-dimensional.
        :type y: 2D Numpy array (time, dimension)
        :arg y: The time series y to be analyzed, can be scalar or
            multi-dimensional.
        :type metric: tuple of string
        :arg metric: The metric for measuring distances in phase space
            ("manhattan", "euclidean", "supremum").  Give separately for each
            time series.
        :type normalize: tuple of bool
        :arg normalize: Decide whether to normalize the time series to zero
            mean and unit standard deviation. Give separately for each time
            series.
        :arg number lag: To create a delayed version of the JRP.
        :arg number silence_level: Inverse level of verbosity of the object.
        :type threshold: tuple of number
        :keyword threshold: The recurrence threshold keyword for generating the
            recurrence plot using a fixed threshold.  Give separately for each
            time series.
        :type threshold_std: tuple of number
        :keyword threshold_std: The recurrence threshold keyword for generating
            the recurrence plot using a fixed threshold in units of the time
            series' STD. Give separately for each time series.
        :type recurrence_rate: tuple of number
        :keyword recurrence_rate: The recurrence rate keyword for generating
            the recurrence plot using a fixed recurrence rate. Give separately
            for each time series.
        :type dim: tuple of number
        :keyword dim: The embedding dimension. Give separately for each time
            series.
        :type tau: tuple of number
        :keyword tau: The embedding delay. Give separately for each time
            series.
        """
        #  Check for consistency
        if np.abs(lag) < x.shape[0]:
            if x.shape[0] == y.shape[0]:
                #  Initialize the underlying RecurrencePlot object
                JointRecurrencePlot.__init__(self, x, y, metric, normalize,
                                             lag, silence_level, **kwds)

                #  Set diagonal of JR to zero to avoid self-loops in the joint
                #  recurrence network
                A = self.JR - np.eye((self.N-np.abs(lag)), dtype="int8")

                #  Create a Network object interpreting the recurrence matrix
                #  as the graph adjacency matrix. Joint recurrence networks
                #  are undirected by definition.
                Network.__init__(self, A, directed=False,
                                 silence_level=silence_level)
            else:
                raise ValueError("Both time series x and y need to have the \
                                 same length!")
        else:
            raise ValueError("Delay value (lag) must not exceed length of \
                             time series!")

    def __str__(self):
        """
        Returns a string representation.
        """
        return ("JointRecurrenceNetwork:\n"
                f"{JointRecurrencePlot.__str__(self)}\n"
                f"{Network.__str__(self)}")

    def clear_cache(self):
        """
        Clean up memory by deleting information that can be recalculated from
        basic data.

        Extends the clean up methods of the parent classes.
        """
        #  Call clean up of RecurrencePlot
        JointRecurrencePlot.clear_cache(self)
        #  Call clean up of Network
        Network.clear_cache(self)

    #
    #  Methods to handle recurrence networks
    #
    def set_fixed_threshold(self, threshold):
        """
        Create a joint recurrence network at fixed thresholds.

        :type threshold: tuple of number
        :arg threshold: The threshold. Give for each time series separately.
        """
        #  Set fixed threshold on recurrence plot level
        JointRecurrencePlot.set_fixed_threshold(self, threshold)

        #  Set diagonal of JR to zero to avoid self-loops in the joint
        #  recurrence network
        A = self.JR.copy()
        A.flat[::self.N+1] = 0

        #  Create a Network object interpreting the recurrence matrix as the
        #  graph adjacency matrix. Joint recurrence networks are undirected by
        #  definition.
        Network.__init__(self, A, directed=False,
                         silence_level=self.silence_level)

    def set_fixed_threshold_std(self, threshold_std):
        """
        Create a joint recurrence network at fixed thresholds in units of the
        standard deviation of the time series.

        :type threshold_std: tuple of number
        :arg threshold_std: The threshold in units of standard deviation. Give
            for each time series separately.
        """
        #  Set fixed threshold on recurrence plot level
        JointRecurrencePlot.set_fixed_threshold_std(self, threshold_std)

        #  Set diagonal of JR to zero to avoid self-loops in the joint
        #  recurrence network
        A = self.JR.copy()
        A.flat[::self.N+1] = 0

        #  Create a Network object interpreting the recurrence matrix as the
        #  graph adjacency matrix. Joint recurrence networks are undirected by
        #  definition.
        Network.__init__(self, A, directed=False,
                         silence_level=self.silence_level)

    def set_fixed_recurrence_rate(self, recurrence_rate):
        """
        Create a joint recurrence network at fixed link densities (recurrence
        rates).

        :type recurrence_rate: tuple of number
        :arg recurrence_rate: The link density / recurrence rate.
            Give for each time series separately.
        """
        #  Set fixed recurrence rate on recurrence plot level
        JointRecurrencePlot.set_fixed_recurrence_rate(self, recurrence_rate)

        #  Set diagonal of JR to zero to avoid self-loops in the joint
        #  recurrence network
        A = self.JR.copy()
        A.flat[::self.N+1] = 0

        #  Create a Network object interpreting the recurrence matrix as the
        #  graph adjacency matrix. Joint recurrence networks are undirected by
        #  definition.
        Network.__init__(self, A, directed=False,
                         silence_level=self.silence_level)
